import numpy as np

class Selectivity():
    """
    Wrapper for selectivity data storage. Selectivities are stored as numpy.ndarray of dictionaries, containing name of compounds and corresponding calculated selectivities. The list is parallel to the list of corresponding temperatures.
    """

    def __init__(self, temperatures:list[float]|np.ndarray[float, np.dtype], selectivities:list[dict[str,float]]|np.ndarray[dict[str,float], np.dtype], sample_name:str|None):
        """
        Assigns parameters to instance variables, converting lists to numpy.ndarrays.

        parameters
        ----------
        temperatures:list[float]|numpy.ndarray[float]
            list of temperatures of catalytic reaction at which measurements were done
        selectivities:list[dict[str,float]]|numpy.ndarray[dict[str,float]]
            list of dictionaries with selectivities parallel to temperatures list. Selectivities are stored in a format {<compound>:<selectivity>}
                compound:str
                    chemical formula of compound
                selectivity:float
                    selectivity of catalyst to this compound
        sample_name:str|None
            name of sample
        """
        self.temperatures = np.array(temperatures)
        self.selectivities = np.array(selectivities)
        self.sample_name = sample_name

    def __str__(self) -> str:
        """
        Get string representation of selectivities in a form of table:

        Temperature<tab>Selectivity

        returns
        -------
        string:str
            string representation of selectivities
        """
        sorted_selectivities = self.get_sorted()
        c_l = []
        for compound in sorted_selectivities.get_selectivities()[0]:
            c_l.append(compound)
        header = f'Sample\t{self.sample_name}\n\nTemperature'
        for compound in c_l:
            header = header + f'\t{compound}'
        header = header + '\n'
        data = ''
        for temperature in sorted_selectivities.get_temperatures():
            data = data + f'{temperature}'
            for compound in c_l:
                data = data + f'\t{sorted_selectivities.get_selectivity(compound, temperature)}'
            data = data + '\n'
        string = header + data
        return string

    def get_temperatures(self) -> np.ndarray[float, np.dtype]:
        """
        Get list of temperatures of catalytic reaction at which measurements were done

        returns
        -------
        temperatures:numpy.ndarray[float]
            list of temperatures
        """
        return self.temperatures

    def get_selectivities(self) -> np.ndarray[dict[str,float], np.dtype]:
        """
        Get list of selectivities calculated by the program

        returns
        -------
        selectivities:numpy.ndarray[dict[str,float]]
            list of selectivities
        """
        return self.selectivities

    def get_selectivity(self, compound:str, temperature:float) -> float:
        """
        Get selectivity of catalyst to compound at temperature of catalytic reaction provided as parameter to the method

        parameters
        ----------
        compound:str
            compound for which selectivity to return
        temperature:float
            temperature of catalytic reaction

        returns
        -------
        selectivity:float
            selectivity to specified compound
        """
        return self.get_selectivities()[self.temperatures==temperature][0][compound]

    def get_sorted(self) -> 'Selectivity':
        """
        Get selectivity object with temperatures and selectivities lists sorted in parallel based on temperatures

        returns
        -------
        selectivity:Selectivity
            sorted selectivity
        """
        zipped_lists = zip(self.get_temperatures(), self.get_selectivities(), strict=True)
        sorted_pairs = sorted(zipped_lists)
        tuples = zip(*sorted_pairs)
        sorted_temperatures, sorted_selectivities = [list(tuple) for tuple in tuples]
        return Selectivity(sorted_temperatures, sorted_selectivities, self.sample_name)

    def get_selectivities_at(self, temperature:float) -> dict[str,float]:
        """
        Get selectivities at temperature of catalytic reaction provided as parameter to the method

        parameters
        ----------
        temperature:float
            temperature of catalytic reaction

        returns
        -------
        selectivities:dict[str,float]
            dictionary of selectivities at specified temperature
        """
        return self.get_selectivities()[self.get_temperatures() == temperature][0]



import propar

from pycatalicism.mass_flow_controller.mfc_exceptions import MFCConnectionException
from pycatalicism.mass_flow_controller.mfc_exceptions import MFCStateException
from pycatalicism.mass_flow_controller.bronkhorst_mfc_calibration import BronkhorstMFCCalibration
import pycatalicism.mass_flow_controller.mass_flow_controller_logging as mfc_logging

class BronkhorstF201CV():
    """
    Class represents Bronkhorst F201CV mass flow controller.
    """

    def __init__(self, serial_address:str, serial_id:str, calibrations:dict[int, BronkhorstMFCCalibration]):
        """
        Initializes instance variables, registers logger.

        parameters
        ----------
        serial_address:str
            serial address of this controller (COM or /dev/ttyUSB)
        serial_id:str
            serial number of this controller (e.g. M111202123A, see documentation)
        calibrations:dict[int, BronkhorstMFCCalibration]
            set of calibrations written in the memory of the device with corresponding calibration number. NB: calibration number is for propar protocol, numbering is started from 0, so fluid1 has 0th number
        """
        self._serial_address = serial_address
        self._serial_id = serial_id
        self._calibrations = calibrations
        self._current_calibration = None
        self._connected = False
        self._logger = mfc_logging.get_logger(self.__class__.__name__)

    def connect(self):
        """
        Connect to mass flow controller. Method requests serial number of the device and compares it with the one, provided at object creation. Also method requests current calibration number. This method must be run before any other method of the class.

        raises
        ------
        MFCConnectionException
            if wrong serial id was received from the device
        """
        self._logger.info(f'Connecting to mass flow controller {self._serial_id}')
        self._propar_instrument = propar.instrument(comport=self._serial_address)
        serial_id_response = self._propar_instrument.readParameter(dde_nr=92)
        self._logger.log(5, f'{serial_id_response = }')
        if not serial_id_response == self._serial_id:
            raise MFCConnectionException(f'Wrong serial {serial_id_response} was received from the device {self._serial_id}')
        self._current_calibration = self._propar_instrument.readParameter(dde_nr=24)
        self._logger.log(5, f'{self._current_calibration = }')
        self._logger.info(f'Current calibration: {self._calibrations[self._current_calibration]}')
        self._connected = True

    def set_flow_rate(self, flow_rate:float):
        """
        Sets flow rate to specified value.

        parameters
        ----------
        flow_rate:float
            flow rate in the units of mass flow controller

        raises
        ------
        MFCStateException
            if mass flow controller is not connected
        """
        if not self._connected:
            raise MFCStateException(f'Mass flow controller {self._serial_id} is not connected!')
        if flow_rate > self._calibrations[self._current_calibration].get_max_flow_rate():
            self._logger.warning('Cannot set flow rate bigger than maximum value in calibraion! Flow rate will be set to maximum value.')
            flow_rate = self._calibrations[self._current_calibration].get_max_flow_rate()
        self._logger.info(f'Setting flow rate to {flow_rate} nml/min')
        percent_setpoint = flow_rate * 100 / self._calibrations[self._current_calibration].get_max_flow_rate()
        self._logger.log(5, f'{percent_setpoint = }')
        propar_setpoint = int(percent_setpoint * 32000 / 100)
        self._logger.log(5, f'{propar_setpoint = }')
        self._propar_instrument.setpoint = propar_setpoint

    def set_calibration(self, calibration_num:int):
        """
        Set calibration to specified calibration number.

        parameters
        ----------
        calibration_num:int
            the number of calibration. NB: the numbering is started from 0, so calibration for fluid1 has 0th order.

        raises
        ------
        MFCStateException
            if mass flow controller is not connected
        """
        if not self._connected:
            raise MFCStateException(f'Mass flow controller {self._serial_id} is not connected!')
        self._logger.info(f'Setting calibration to {self._calibrations[calibration_num]}')
        self._propar_instrument.writeParameter(dde_nr=24, data=calibration_num)
        self._current_calibration = calibration_num

    def get_flow_rate(self) -> float:
        """
        Get current flow rate from mass flow controller.

        returns
        -------
        flow_rate:float
            flow rate in the units of mass flow controller

        raises
        ------
        MFCStateException
            if mass flow controller is not connected
        """
        if not self._connected:
            raise MFCStateException(f'Mass flow controller {self._serial_id} is not connected!')
        self._logger.info('Reading current flow rate')
        flow_rate_propar = self._propar_instrument.measure
        self._logger.log(5, f'{flow_rate_propar = }')
        if flow_rate_propar is None:
            raise MFCConnectionException(f'Failed to get flow rate from the instrument {self._serial_id}')
        flow_rate_percent = flow_rate_propar / 32000.0
        self._logger.log(5, f'{flow_rate_percent = }')
        flow_rate = flow_rate_percent * self._calibrations[self._current_calibration].get_max_flow_rate()
        self._logger.log(5, f'{flow_rate = }')
        return flow_rate

    def get_calibration(self) -> BronkhorstMFCCalibration:
        """
        Get current calibration set in mass flow controller.

        returns
        -------
        calibration:BronkhorstMFCCalibration
            wrapper with calibration data

        raises
        ------
        MFCStateException
            if mass flow controller is not connected
        """
        if not self._connected:
            raise MFCStateException(f'Mass flow controller {self._serial_id} is not connected!')
        self._logger.info('Getting current calibration')
        calibration = self._calibrations[self._current_calibration]
        self._logger.log(5, f'{calibration = }')
        return calibration



import threading
import serial
import struct
import time

import pycatalicism.furnace.furnace_logging as furnace_logging
from pycatalicism.furnace.furnace_exceptions import FurnaceProtocolException
from pycatalicism.furnace.furnace_exceptions import FurnaceConnectionException

class OwenProtocol():
    """
    Class represents simplified Owen Protocol. It does not support all parameter types as well as ignores some peculiarities of the protocol.
    """

    def __init__(self, address:int, port:str, baudrate:int, bytesize:int, parity:str, stopbits:float, timeout:float, write_timeout:float|None, rtscts:bool, request_trials:int=3):
        """
        Initialize instance variables, gather read/write lock and register logger.

        parameters
        ----------
        address:int
            address of the controller to comunicate with. Must be the same as on the controller.
        port:str
            serial port controller is connected to
        baudrate:int
            baudrate to use during communication via serial port. Must be the same as on the controller.
        bytesize:int
            bytesize to use during communication via serial port. Must be the same as on the controller.
        parity:str
            parity to use during communication via serial port. Must be the same as on the controller.
        stopbits:float
            stopbits to use during communication via serial port. Must be the same as on the controller.
        timeout:float
            timeout to use during communication via serial port. See pyserial for details.
        write_timeout:float|None
            write timeout to use during communication via serial port. See pyserial for details.
        rtscts:bool
            enable hardware flow control. See pyserial for details.
        request_trials:int (default: 3)
            how many times to try connecting to the controller before exception is thrown
        """
        self._address = address
        self._port = port
        self._baudrate = baudrate
        self._bytesize = bytesize
        self._parity = parity
        self._stopbits = stopbits
        self._timeout = timeout
        self._rtscts = rtscts
        self._write_timeout = write_timeout
        self._read_write_lock = threading.Lock()
        self._request_trials = request_trials
        self._logger = furnace_logging.get_logger(self.__class__.__name__)

    ## Public interface ##

    def request_string(self, parameter:str) -> str:
        """
        Request parameter of type ASCII string from the controller.

        parameters
        ----------
        parameters:str
            parameter id

        returns
        -------
        string:str
            parameter value returned by the controller

        raises
        ------
        FurnaceConnectionException
            if several trials was unsuccessful to get the parameter value
        """
        message = self._pack_message(command=parameter, is_request=True, data=None)
        count = 0
        while True:
            self._logger.debug(f'Requesting string parameter "{parameter}". Trial #{count}.')
            try:
                parameter_data_bytes = self._get_parameter_data_bytes(message=message)
                self._logger.log(5, f'{parameter_data_bytes = }')
                string = self._decrypt_string(parameter_data_bytes)
                self._logger.log(5, f'{string = }')
                break
            except FurnaceProtocolException as ex:
                if count == self._request_trials-1:
                    raise FurnaceConnectionException(ex)
                else:
                    count += 1
                    time.sleep(1)
                self._logger.warning(f'Connection problems were encountered: {ex}, trying to repeat. Trial #{count}')
        return string

    def request_PIC(self, parameter:str) -> float:
        """
        Request parameter value of PIC type.

        parameters
        ----------
        parameter:str
            parameter id

        returns
        -------
        pic:float
            parameter valie returned by the controller

        raises
        ------
        FurnaceConnectionException
            if several trials were unsuccessful to get the parameter value
        """
        message = self._pack_message(command=parameter, is_request=True, data=None)
        count = 0
        while True:
            self._logger.debug(f'Requesting PIC parameter "{parameter}". Trial #{count}.')
            try:
                parameter_data_bytes = self._get_parameter_data_bytes(message)
                self._logger.log(5, f'{parameter_data_bytes = }')
                pic = self._decrypt_PIC(parameter_data_bytes)
                self._logger.log(5, f'{pic = }')
                break
            except FurnaceProtocolException as ex:
                if count == self._request_trials-1:
                    raise FurnaceConnectionException(ex)
                else:
                    count += 1
                    time.sleep(1)
                self._logger.warning(f'Connection problems were encountered: {ex}, trying to repeat. Trial #{count}')
        return pic

    def send_PIC(self, parameter:str, value:float):
        """
        Set parameter value of type PIC to specified value.

        parameters
        ----------
        parameters:str
            parameter id
        value:float
            new value to set the parameter to

        raises
        ------
        FurnaceConnectionException
            if several trials were unsuccessful to set the new value
        """
        data = self._float_to_PIC(value=value)
        message = self._pack_message(command=parameter, is_request=False, data=data)
        count = 0
        while True:
            self._logger.debug(f'Sending PIC parameter "{parameter}". Trial #{count}.')
            try:
                self._change_parameter_value(message=message)
                break
            except FurnaceProtocolException as ex:
                if count == self._request_trials-1:
                    raise FurnaceConnectionException(ex)
                else:
                    count += 1
                    time.sleep(1)

    def send_unsigned_byte(self, parameter:str, value:int):
        """
        Set parameter value of type unsigned byte to the new value.

        parameters
        ----------
        parameter:str
            parameter id
        value:int
            new value to set the parameter to

        raises
        ------
        FurnaceConnectionException
            if several trials were unsuccessful to set the new value.
        """
        data = self._int_to_unsigned_byte(value=value)
        message = self._pack_message(command=parameter, is_request=False, data=data)
        count = 0
        while True:
            self._logger.debug(f'Sending unsigned byte parameter "{parameter}". Trial #{count}.')
            try:
                self._change_parameter_value(message=message)
                break
            except FurnaceProtocolException as ex:
                if count == self._request_trials-1:
                    raise FurnaceConnectionException(ex)
                else:
                    count += 1
                    time.sleep(1)

    ## Top level i/o ##

    def _change_parameter_value(self, message:str):
        """
        Send encoded tetrad-to-ASCII message with parameter change request to the controller. Get receipt and check if it is ok.

        parameters
        ----------
        message:str
            encoded tetrad-to-ASCII message

        raises
        ------
        FurnaceProtocolException
            if receipt is not ok
        """
        with self._read_write_lock:
            self._write_message(message)
            receipt = self._read_message()
        if not self._receipt_is_ok(receipt=receipt, message=message):
            raise FurnaceProtocolException('Got wrong receipt from device!')

    def _get_parameter_data_bytes(self, message:str) -> list[int]:
        """
        Send encoded tetrad-to-ASCII message with the parameter value request to the controller. Get response, unpack it and check if crc is ok.

        parameters
        ----------
        message:str
            Message encrypted in tetrad-to-ASCII form according to owen protocol to be sent to the device.

        returns
        -------
        data:list[int]
            requested parameter value in byte form

        raises
        ------
        FurnaceProtocolException
            if crc checksum is not ok or if there were no data in controller's response.
        """
        with self._read_write_lock:
            self._write_message(message)
            response = self._read_message()
        address, flag_byte, response_hash, data, crc = self._unpack_message(response)
        if not self._crc_is_ok(address, flag_byte, response_hash, data, crc):
            raise FurnaceProtocolException(f'Wrong CRC in response message!')
        if data is None:
            raise FurnaceProtocolException('Did not get any data in response message')
        return data

    ## Bottom level i/o ##

    def _write_message(self, message:str):
        """
        Writes enctypted message over serial port.

        parameters
        ----------
        message:str
            Encrypted message to be sent to the device
        """
        with serial.Serial(port=self._port, baudrate=self._baudrate, bytesize=self._bytesize, parity=self._parity, stopbits=self._stopbits, timeout=self._timeout, rtscts=self._rtscts, write_timeout=self._write_timeout) as ser:
            self._logger.log(5, f'Writing message: {bytes(message, encoding="ascii")}')
            ser.write(bytes(message, encoding='ascii'))

    def _read_message(self) -> str:
        """
        Reads message from the device over serial port.

        returns
        -------
        message:str
            Encrypted message received from the device

        raises
        ------
        FurnaceProtocolException
            If message does not contain proper start and stop markers
        """
        with serial.Serial(port=self._port, baudrate=self._baudrate, bytesize=self._bytesize, parity=self._parity, stopbits=self._stopbits, timeout=self._timeout, rtscts=self._rtscts, write_timeout=self._write_timeout) as ser:
            message = ''
            for i in range(44):
                self._logger.log(5, f'Reading byte #{i}')
                byte = ser.read().decode()
                self._logger.log(5, f'Read byte: {byte}')
                message = message + byte
                if byte == chr(0x0d):
                    break
            self._logger.debug(f'Got message: {message = }')
        if len(message) == 0:
            raise FurnaceProtocolException('Empty message was received from device')
        if message[0] != chr(0x23) or message[-1] != chr(0x0d):
            raise FurnaceProtocolException(f'Unexpected format of message got from device: {message}')
        return message

    ## Encrypt data to bytes ##

    def _float_to_PIC(self, value:float) -> list[int]:
        """
        Encrypt float value to PIC bytes according to owen protocol. Encrypted bytes can be sent in data block of the message.

        parameters
        ----------
        value:float
            Value to encrypt

        returns
        -------
        pic_bytes:list[int]
            List of 3 bytes of encrypted value
        """
        pic_bytes = []
        ieee = struct.pack('>f', value)
        self._logger.log(5, f'{[b for b in ieee] = }')
        for i in range(3):
            pic_bytes.append(ieee[i])
        self._logger.debug(f'{pic_bytes = }')
        return pic_bytes

    def _int_to_unsigned_byte(self, value:int) -> list[int]:
        """
        Encrypt int value to unsigned byte according to owen protocol. Encrypted bytes can be sent in data block of the message

        parameters
        ----------
        value:int
            value to convert to byte

        returns
        -------
        unsigned_byte:list[int]
            list of 1 byte with converted value

        raises
        ------
        FurnaceProtocolException
            if value > 255 or value < 0
        """
        if value > 255 or value < 0:
            raise FurnaceProtocolException(f'Got wrong value to convert to unsigned byte: {value}')
        unsigned_bytes = [value]
        self._logger.debug(f'{unsigned_bytes = }')
        return unsigned_bytes

    def _str_to_ASCII(self, value:str) -> list[int]: # NOT USED IN CURRENT IMPLEMENTATION OF THE PROTOCOL
        """
        Encrypt string value to ASCII bytes according to owen protocol. Encrypted bytes can be sent in data block of the message.

        parameters
        ----------
        value:str
            String to be encrypted and sent to the device according to owen protocol

        returns
        -------
            List of ASCII bytes to be sent to the device

        raises
        ------
        FurnaceProtocolException
            If non-ASCII character was encountered in the value
        """
        ascii_bytes = []
        for ch in value[::-1]:
            if ord(ch) > 127:
                raise FurnaceProtocolException(f'Non ASCII character was met in value: {ch}')
            ascii_bytes.append(ord(ch))
        return ascii_bytes

    ## Decrypt data from bytes ##

    def _decrypt_PIC(self, data:list[int]) -> float:
        """
        Decrypt float value from PIC bytes received from the device

        parameters
        ----------
        data:list[int]
            List of 3 bytes received from the device

        returns
        -------
        pic:float
            Float value decrypted according to owen protocol

        raises
        ------
        FurnaceProtocolException
            If data is None or if size of data list is greater than 3 bytes
        """
        if data is None:
            raise FurnaceProtocolException('Cannot decrypt empty data')
        if len(data) > 3:
            raise FurnaceProtocolException('Unexpected size of data to convert to PIC float')
        data_str = b''
        for b in data:
            data_str = data_str + b.to_bytes(1, 'big')
        data_str = data_str + int(0).to_bytes(1, 'big')
        pic = struct.unpack('>f', data_str)[0]
        return pic

    def _decrypt_string(self, data:list[int]|None) -> str:
        """
        Decrypt string message from data bytes received from the device.

        parameters
        ----------
        data:list[int]
            List of bytes with encrypted string value

        returns
        -------
        string:str
            Decrypted string value

        raises
        ------
        FurnaceProtocolException
            If data is None
        """
        self._logger.log(5, f'{data = }')
        if data is None:
            raise FurnaceProtocolException('Cannot decrypt empty data!')
        string = ''
        for data_byte in data[::-1]:
            string = string + chr(data_byte)
        return string

    ## Message packing/unpacking ##

    def _pack_message(self, command:str, is_request:bool, data:list[int]|None) -> str:
        """
        Prepares ASCII message in tetrad-to-ASCII form according to owen protocol. Methods gets command id, retreives command hash and encrypts message.

        parameters
        ----------
        command:str
            Parameter name to get value of from the device

        returns
        -------
        message_ascii:str
            Encrypted in tetrad-to-ASCII form according to owen protocol message to be sent to the device
        """
        command_id = self._get_command_id(command)
        command_hash = self._get_command_hash(command_id)
        data_length = 0 if data is None else len(data)
        message_ascii = self._encrypt_tetrad_to_ascii(address=self._address, request=is_request, data_length=data_length, command_hash=command_hash, data=data)
        return message_ascii

    def _get_command_id(self, command:str) -> list[int]:
        """
        Encrypt command name to command id according to owen protocol.

        parameters
        ----------
        command:str
            String representation of command

        returns
        -------
        command_id:list[int]
            List of 4 bytes representing encrypted command id according to owen protocol

        raises
        ------
        FurnaceProtocolException
            If illegal char encountered in command name or if command id has more than 4 bytes
        """
        command_id = []
        command_cap = command.upper()
        for i in range(len(command_cap)):
            if command_cap[i] == '.':
                continue
            elif command_cap[i].isdecimal():
                ch_id = ord(command_cap[i]) - ord('0')
            elif command_cap[i].isalpha():
                ch_id = ord(command_cap[i]) - ord('A') + 10
            elif command_cap[i] == '-':
                ch_id = 36
            elif command_cap[i] == '_':
                ch_id = 37
            elif command_cap[i] == '/':
                ch_id = 38
            else:
                raise FurnaceProtocolException(f'Illegal char in command name: {command_cap[i]}')
            ch_id = ch_id * 2
            if i < len(command_cap) - 1 and command_cap[i+1] == '.':
                ch_id = ch_id + 1
            command_id.append(ch_id)
        if len(command_id) > 4:
            raise FurnaceProtocolException('Command ID cannot contain more than 4 characters!')
        if len(command_id) < 4:
            for i in range(4 - len(command_id)):
                command_id.append(78)
        return command_id

    def _get_command_hash(self, command_id:list[int]) -> int:
        """
        Calculates 2 bytes command hash according to owen protocol.

        parameters
        ----------
        command_id:list[int]
            List of 4 bytes encrypted command id

        returns
        -------
        command_hash:int
            2 bytes command hash
        """
        command_hash = 0
        for b in command_id:
            b = b << 1
            b = b & 0xff
            for i in range(7):
                if (b ^ (command_hash >> 8)) & 0x80:
                    command_hash = command_hash << 1
                    command_hash = command_hash ^ 0x8f57
                else:
                    command_hash = command_hash << 1
                command_hash = command_hash & 0xffff
                b = b << 1
                b = b & 0xff
        self._logger.log(5, f'{command_hash = :#x}')
        return command_hash

    def _get_crc(self, message_bytes:list[int]) -> int:
        """
        Calculate CRC check sum for message according to owen protocol.

        parameters
        ----------
        message_bytes:list[int]
            Message bytes containing address byte, flag byte, hash bytes and data bytes encrypted according owen protocol.

        returns
        -------
        crc:int
            2 bytes CRC chack sum according owen protocol.
        """
        crc = 0
        for b in message_bytes:
            b = b & 0xff
            for i in range(8):
                if (b ^ (crc >> 8)) & 0x80:
                    crc = crc << 1
                    crc = crc ^ 0x8f57
                else:
                    crc = crc << 1
                crc = crc & 0xffff
                b = b << 1
                b = b & 0xff
        return crc

    def _encrypt_tetrad_to_ascii(self, address:int, request:bool, data_length:int, command_hash:int, data:list[int]|None) -> str:
        """
        Encrypt message in tetrad-to-ASCII form according to owen protocol.

        parameters
        ----------
        address:int
            Byte with address of the device. Must match the one configured on the device.
        request:bool
            True if message contains request of parameter value.
        data_length:int
            Number of bytes used to encrypt data. Must be in range [0,15]
        command_hash:int
            2 bytes encrypted according to owen protocol command hash.
        data:list[int] or None
            Data to be sent to the device encrypted as byte list according to the owen protocol or None if no data is sent to the device

        returns
        -------
        message:str
            tetrad-to-ASCII encrypted according to owen protocol message

        raises
        ------
        FurnaceProtocolException
            If data length is larger 15 or data_length parameter does not match length of data bytes list or if encrypted message contains wrong characters
        """
        if data_length > 15:
            raise FurnaceProtocolException('Data length cannot be larger than 15')
        if data is None and data_length != 0:
            raise FurnaceProtocolException('data_length parameter cannot be non zero if data is None')
        message_bytes = []
        message_bytes.append(address & 0xff)
        # NB: if address_len is 11b flag_byte must be modified accordingly
        flag_byte = 0
        if request:
            flag_byte = flag_byte | 0b00010000
        flag_byte = flag_byte | data_length
        message_bytes.append(flag_byte & 0xff)
        message_bytes.append((command_hash >> 8) & 0xff)
        message_bytes.append(command_hash & 0xff)
        if data is not None:
            if len(data) > 15:
                raise FurnaceProtocolException('Data length cannot be larger than 15')
            if len(data) != data_length:
                raise FurnaceProtocolException('Length of data bytes list does not match data_length parameter')
            for data_byte in data:
                message_bytes.append(data_byte & 0xff)
        crc = self._get_crc(message_bytes)
        message_bytes.append((crc >> 8) & 0xff)
        message_bytes.append(crc & 0xff)
        message = chr(0x23)
        for byte in message_bytes:
            message = message + chr(((byte >> 4) & 0xf) + 0x47)
            message = message + chr((byte & 0xf) + 0x47)
        message = message + chr(0x0d)
        for ch in message:
            if ch not in ['#', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', '\r']:
                raise FurnaceProtocolException(f'Wrong ASCII message: "{message}"!')
        return message

    def _unpack_message(self, message:str) -> tuple[int, int, int, list[int]|None, int]:
        """
        Decrypt tetrad-to-ASCII encrypted message according to owen protocol into bytes.

        parameters
        ----------
        message:str
            Tetrad-to-ASCII encrypted message received from the device

        returns
        -------
        address:int
            Byte with address parameter
        flag_byte:int
            Byte with enhanced address, request and data length bits
        response_hash:int
            2 bytes hash of command id
        data:list[int] or None
            List of bytes with data encrypted according to owen protocol or None if data length is 0
        crc:int
            2 bytes CRC check sum received in message

        raises
        ------
        FurnaceProtocolException
            If received message does not contain proper start and stop markers
        """
        if message[0] != chr(0x23) or message[-1] != chr(0x0d):
            raise FurnaceProtocolException(f'Unexpected format of message from device: {message}')
        message_bytes = []
        for i in range(1, len(message) - 1, 2):
            first_tetrad = (ord(message[i]) - 0x47) & 0xf
            second_tetrad = (ord(message[i+1]) - 0x47) & 0xf
            self._logger.log(5, f'ASCII letter #{i} = {message[i]}')
            self._logger.log(5, f'{first_tetrad = :#b}')
            self._logger.log(5, f'{second_tetrad = :#b}')
            byte = ((first_tetrad << 4) | second_tetrad) & 0xff
            message_bytes.append(byte)
        self._logger.log(5, f'{len(message_bytes) = }')
        self._logger.log(5, f'{message_bytes = }')
        address = message_bytes[0]
        self._logger.log(5, f'{address = }')
        flag_byte = message_bytes[1]
        self._logger.log(5, f'{flag_byte = :#b}')
        response_hash = ((message_bytes[2] << 8) | message_bytes[3]) & 0xffff
        self._logger.log(5, f'{response_hash = :#x}')
        data_length = flag_byte & 0b1111
        self._logger.log(5, f'{data_length = }')
        if data_length != 0:
            data = []
            for i in range(data_length):
                data.append(message_bytes[4 + i])
        else:
            data = None
        crc = ((message_bytes[4+data_length] << 8) | message_bytes[4+data_length+1]) & 0xffff
        return (address, flag_byte, response_hash, data, crc)

    ## Checking ##

    def _receipt_is_ok(self, receipt:str, message:str) -> bool:
        """
        Checks whether receipt received from the device is correct.

        parameters
        ----------
        receipt:str
            Receipt message received from the device in enctypted form
        message:str
            Message that was sent to the device in enctypted form

        returns
        -------
        receipt_is_ok:bool
            True if receipt is correct
        """
        address, flag_byte, response_hash, data, crc = self._unpack_message(receipt)
        self._logger.log(5, f'Receipt address: {address}')
        self._logger.log(5, f'Receipt flag_byte: {flag_byte:#b}')
        self._logger.log(5, f'Receipt response_hash: {response_hash:#x}')
        self._logger.log(5, f'Receipt data: {data}')
        self._logger.log(5, f'Receipt crc: {crc}')
        self._logger.log(5, f'Receipt crc is ok: {self._crc_is_ok(address, flag_byte, response_hash, data, crc)}')
        new_flag_tetrad = (ord(message[3]) - 0x47) & 0b1110
        new_flag_chr = chr((new_flag_tetrad & 0xf) + 0x47)
        message_without_request = ''
        for i in range(len(message)):
            if i == 3:
                message_without_request = message_without_request + new_flag_chr
            else:
                message_without_request = message_without_request + message[i]
        receipt_is_ok = message_without_request == receipt
        self._logger.debug(f'Receipt is ok: {receipt_is_ok}')
        return receipt_is_ok

    def _crc_is_ok(self, address:int, flag_byte:int, response_hash:int, data:list[int]|None, crc_to_check:int) -> bool:
        """
        Checks CRC check sum of received message.

        parameters
        ----------
        address:int
            Address byte
        flag_byte:int
            Flag byte
        response_hash:int
            2 byte response command hash
        data:list[int] or None
            List of bytes with enctypted data or None if data bytes were empty
        crc_to_check:int
            2 bytes CRC check sum that was received in the message

        returns
        -------
        crc_is_ok:bool
            True if CRC check sum is correct
        """
        message_bytes = []
        message_bytes.append(address)
        message_bytes.append(flag_byte)
        message_bytes.append((response_hash >> 8) & 0xff)
        message_bytes.append(response_hash & 0xff)
        if data is not None:
            for data_byte in data:
                message_bytes.append(data_byte & 0xff)
        crc = self._get_crc(message_bytes)
        crc_is_ok = crc == crc_to_check
        self._logger.debug(f'{crc_is_ok = }')
        return crc_is_ok



import logging

# Dictionary of logging levels of corresponding classes
logging_levels = {
                'ChromatecControlPanelModbus'               	:		logging.INFO,
                'ChromatecAnalyticModbus'						:		logging.INFO,
                'ChromatecCrystal5000'							:		logging.INFO,
                'pycatalicism.chromatograph.modbus_converter'	:		logging.INFO,
                }



class Point():
    """
    Wrapper class to store 2D data point. Used by non-blocking plotter to send and receive data through multiprocessing pipe.
    """

    def __init__(self, x:float, y:float|None, label:str):
        """
        Initialize instance variables.

        parameters
        ----------
        x:float
            x coordinate
        y:float|None
            y coordinate, None means that this is 1D data
        label:str
            label to associate with the data
        """
        self._x = x
        self._y = y
        self._label = label

    def get_x(self) -> float:
        """
        Get x value.

        returns
        -------
        x:float
            x value
        """
        return self._x

    def get_y(self) -> float|None:
        """
        Get y value.

        returns
        -------
        y:float
            y value
        """
        return self._y

    def get_label(self) -> str:
        """
        Get label associated with data.

        returns
        -------
        label:str
            label associated with data
        """
        return self._label



class BronkhorstMFCCalibration():
    """
    Class is a wrapper to mass flow controller calibration
    """

    def __init__(self, max_flow_rate:float, gas:str, p_in:float, p_out:float):
        """
        Initializes instance variables

        parameters
        ----------
        max_flow_rate:float
            Maximum flow rate in calibration in the units of mass flow controller
        gas:str
            Gas name
        p_in:float
            Pressure at the inlet of mass flow controller used in calibration
        p_out:float
            Pressure at the outlet of mass flow controller used in calibration
        """
        self._max_flow_rate = max_flow_rate
        self._gas = gas
        self._p_in = p_in
        self._p_out = p_out

    def __str__(self) -> str:
        """
        Get string representation of the object.

        returns
        -------
        string:str
            String representation of the calibration
        """
        string = f'[gas: {self._gas}, p_in: {self._p_in}, p_out: {self._p_out}, flow rate: {self._max_flow_rate} nml/min]'
        return string

    def get_max_flow_rate(self) -> float:
        """
        Get maximum flow rate of calibration in the units of mass flow controller

        returns
        -------
        max_flow_rate:float
            maximum flow rate
        """
        return self._max_flow_rate



class MFCConnectionException(Exception):
    """
    Class represents exceptions related to connection errors
    """

class MFCStateException(Exception):
    """
    Class represents exceptions related to wrong state of the device
    """



#!/usr/bin/python

"""
Module is a start point for the program. It parses arguments provided by user as command line arguments and executes corresponding functions.
"""

import argparse
import time
import importlib
import importlib.util
import sys
from pathlib import Path
from datetime import date
import types

import pycatalicism.calc.calc as calc
import pycatalicism.config as config
from pycatalicism.calc.calculatorexception import CalculatorException
from pycatalicism.furnace.owen_protocol import OwenProtocol
from pycatalicism.furnace.owen_tmp101 import OwenTPM101
from pycatalicism.chromatograph.chromatec_control_panel_modbus import ChromatecControlPanelModbus
from pycatalicism.chromatograph.chromatec_analytic_modbus import ChromatecAnalyticModbus
from pycatalicism.chromatograph.chromatec_analytic_modbus import ChromatogramPurpose
from pycatalicism.chromatograph.chromatec_crystal_5000 import ChromatecCrystal5000
from pycatalicism.chromatograph.chromatec_control_panel_modbus import WorkingStatus
from pycatalicism.mass_flow_controller.bronkhorst_f201cv import BronkhorstF201CV
from pycatalicism.plotters.process_plotter import DataCollectorPlotter

def calculate(args:argparse.Namespace):
    """
    Calculate conversion and/or selectivity (depending on --conversion/--selectivity flag provided by user) vs. temperature for CO oxidation or CO2 hydrogenation reactions, print results to console and export them if path to export directory was provided by user. Plot corresponding graphs if --show-plot argument was provided by user and export them if export directory was provided.
    """
    parser_type = config.raw_data_parser_type
    try:
        calc.calculate(input_data_path=args.input_data_path, initial_data_path=args.initial_data_path, reaction=args.reaction, parser_type=parser_type, calculate_conversion=args.conversion, calculate_selectivity=args.selectivity, products_basis=args.products_basis, output_data_path=args.output_data, show_plot=args.show_plot, output_plot_path=args.output_plot, sample_name=args.sample_name)
    except CalculatorException:
        print('At least one of the flags {--conversion|--selectivity} must be provided to the program')

def furnace_set_temperature(args:argparse.Namespace):
    """
    Set furnace temperature to specified value
    """
    furnace_controller_protocol = OwenProtocol(address=config.furnace_address, port=config.furnace_port, baudrate=config.furnace_baudrate, bytesize=config.furnace_bytesize, parity=config.furnace_parity, stopbits=config.furnace_stopbits, timeout=config.furnace_timeout, write_timeout=config.furnace_write_timeout, rtscts=config.furnace_rtscts)
    furnace_controller = OwenTPM101(device_name=config.furnace_device_name, owen_protocol=furnace_controller_protocol)
    temperature = float(args.temperature)
    furnace_controller.connect()
    furnace_controller.set_temperature(temperature)
    if temperature == 0:
        furnace_controller.set_temperature_control(False)
    else:
        furnace_controller.set_temperature_control(True)

def furnace_print_temperature(args:argparse.Namespace):
    """
    Print current temperature to console
    """
    furnace_controller_protocol = OwenProtocol(address=config.furnace_address, port=config.furnace_port, baudrate=config.furnace_baudrate, bytesize=config.furnace_bytesize, parity=config.furnace_parity, stopbits=config.furnace_stopbits, timeout=config.furnace_timeout, write_timeout=config.furnace_write_timeout, rtscts=config.furnace_rtscts)
    furnace_controller = OwenTPM101(device_name=config.furnace_device_name, owen_protocol=furnace_controller_protocol)
    furnace_controller.connect()
    temperature = furnace_controller.get_temperature()
    print(f'Current temperature is {temperature}°C')

def _initialize_chromatograph() -> ChromatecCrystal5000:
    """
    Initialize modbus objects and chromatograph object with parameters in config.py file, connect to chromatograph.

    returns
    -------
    chromatograph:ChromatecCrystal5000
        chromatograph used for analysis
    """
    control_panel_modbus = ChromatecControlPanelModbus(modbus_id=config.control_panel_modbus_id, working_status_input_address=config.working_status_input_address, step_time_input_address=config.step_time_input_address, connection_status_input_address=config.connection_status_input_address, method_holding_address=config.method_holding_address, chromatograph_command_holding_address=config.chromatograph_command_holding_address, application_command_holding_address=config.application_command_holding_address)
    analytic_modbus = ChromatecAnalyticModbus(modbus_id=config.analytic_modbus_id, sample_name_holding_address=config.sample_name_holding_address, chromatogram_purpose_holding_address=config.chromatogram_purpose_holding_address, sample_volume_holding_address=config.sample_volume_holding_address, sample_dilution_holding_address=config.sample_dilution_holding_address, operator_holding_address=config.operator_holding_address, column_holding_address=config.column_holding_address, lab_name_holding_address=config.lab_name_holding_address)
    chromatograph = ChromatecCrystal5000(control_panel_modbus, analytic_modbus, config.methods)
    chromatograph.connect()
    return chromatograph

def chromatograph_set_method(args:argparse.Namespace):
    """
    Set chromatograph instrument method
    """
    chromatograph = _initialize_chromatograph()
    chromatograph.set_method(method=args.method)

def chromatograph_start_analysis(args:argparse.Namespace):
    """
    Start chromatograph analysis
    """
    chromatograph = _initialize_chromatograph()
    chromatograph.start_analysis()

def chromatograph_set_passport(args:argparse.Namespace):
    """
    Set values of chromatogram passport. Should be run after analysis is complete.
    """
    chromatograph = _initialize_chromatograph()
    if args.purpose == 'analysis':
        purpose = ChromatogramPurpose.ANALYSIS
    elif args.purpose == 'graduation':
        purpose = ChromatogramPurpose.GRADUATION
    else:
        raise Exception(f'Unknown chromatogram purpose: {args.purpose}')
    chromatograph.set_passport(name=args.name, volume=float(args.volume), dilution=float(args.dilution), purpose=purpose, operator=args.operator, column=args.column, lab_name=args.lab_name)

def mfc_set_flow_rate(args:argparse.Namespace):
    """
    Set flow rate of mfc to specified value.
    """
    mfc_He = BronkhorstF201CV(serial_address=config.mfc_He_serial_address, serial_id=config.mfc_He_serial_id, calibrations=config.mfc_He_calibrations)
    mfc_CO2 = BronkhorstF201CV(serial_address=config.mfc_CO2_serial_address, serial_id=config.mfc_CO2_serial_id, calibrations=config.mfc_CO2_calibrations)
    mfc_H2 = BronkhorstF201CV(serial_address=config.mfc_H2_serial_address, serial_id=config.mfc_H2_serial_id, calibrations=config.mfc_H2_calibrations)
    gas = args.gas
    flow_rate = float(args.flow_rate)
    if gas == 'He':
        mfc_He.connect()
        mfc_He.set_flow_rate(flow_rate)
    elif gas == 'CO2' or gas == 'O2':
        mfc_CO2.connect()
        mfc_CO2.set_flow_rate(flow_rate)
    elif gas == 'H2' or gas == 'CO' or gas == 'CH4':
        mfc_H2.connect()
        mfc_H2.set_flow_rate(flow_rate)
    else:
        raise Exception(f'Unknown gas {gas}!')

def mfc_set_calibration(args:argparse.Namespace):
    """
    Set calibration for mass flow controller.
    """
    mfc_He = BronkhorstF201CV(serial_address=config.mfc_He_serial_address, serial_id=config.mfc_He_serial_id, calibrations=config.mfc_He_calibrations)
    mfc_CO2 = BronkhorstF201CV(serial_address=config.mfc_CO2_serial_address, serial_id=config.mfc_CO2_serial_id, calibrations=config.mfc_CO2_calibrations)
    mfc_H2 = BronkhorstF201CV(serial_address=config.mfc_H2_serial_address, serial_id=config.mfc_H2_serial_id, calibrations=config.mfc_H2_calibrations)
    gas = args.gas
    calibration_num = int(args.calibration_number)
    if gas == 'He':
        mfc_He.connect()
        mfc_He.set_calibration(calibration_num=calibration_num-1)
    elif gas in ['CO2', 'O2']:
        mfc_CO2.connect()
        mfc_CO2.set_calibration(calibration_num=calibration_num-1)
    elif gas in ['H2', 'CO', 'CH4']:
        mfc_H2.connect()
        mfc_H2.set_calibration(calibration_num=calibration_num-1)
    else:
        raise Exception(f'Unknown gas {gas}!')

def mfc_print_flow_rate(args:argparse.Namespace):
    """
    Print current flow rate.
    """
    mfc_He = BronkhorstF201CV(serial_address=config.mfc_He_serial_address, serial_id=config.mfc_He_serial_id, calibrations=config.mfc_He_calibrations)
    mfc_CO2 = BronkhorstF201CV(serial_address=config.mfc_CO2_serial_address, serial_id=config.mfc_CO2_serial_id, calibrations=config.mfc_CO2_calibrations)
    mfc_H2 = BronkhorstF201CV(serial_address=config.mfc_H2_serial_address, serial_id=config.mfc_H2_serial_id, calibrations=config.mfc_H2_calibrations)
    gas = args.gas
    if gas == 'He':
        mfc_He.connect()
        print(f'{mfc_He.get_flow_rate()} nml/min')
    elif gas in ['CO2', 'O2']:
        mfc_CO2.connect()
        print(f'{mfc_CO2.get_flow_rate()} nml/min')
    elif gas in ['H2', 'CO', 'CH4']:
        mfc_H2.connect()
        print(f'{mfc_H2.get_flow_rate()} nml/min')
    else:
        raise Exception(f'Unknown gas {gas}!')

def _import_config(path:Path) -> types.ModuleType:
    """
    """
    config_spec = importlib.util.spec_from_file_location('process_config', path)
    if config_spec is None:
        raise Exception(f'Cannot read config file at {path}')
    config_loader = config_spec.loader
    if config_loader is None:
        raise Exception(f'Cannot read config file at {path}')
    config_module = importlib.util.module_from_spec(config_spec)
    sys.modules['process_config'] = config_module
    config_loader.exec_module(config_module)
    process_config = importlib.import_module('process_config')
    return process_config

def _initialize_furnace_controller() -> OwenTPM101:
    """
    Initialize furnace controller with patameters in config.py file. Connect to furnace controller.

    returns
    -------
    furnace_controller:OwenTPM101
        furnace controller
    """
    furnace_controller_protocol = OwenProtocol(address=config.furnace_address, port=config.furnace_port, baudrate=config.furnace_baudrate, bytesize=config.furnace_bytesize, parity=config.furnace_parity, stopbits=config.furnace_stopbits, timeout=config.furnace_timeout, write_timeout=config.furnace_write_timeout, rtscts=config.furnace_rtscts)
    furnace_controller = OwenTPM101(device_name=config.furnace_device_name, owen_protocol=furnace_controller_protocol)
    furnace_controller.connect()
    return furnace_controller

def _initialize_mass_flow_controllers() -> list[BronkhorstF201CV]:
    """
    Initialize 3 mass flow controllers with parameters in config.py file. Connect to mass flow controllers.

    returns
    -------
    mfcs:list[BronkhorstF201CV]
        list of mass flow controllers
    """
    mfcs = list()
    mfcs.append(BronkhorstF201CV(serial_address=config.mfc_He_serial_address, serial_id=config.mfc_He_serial_id, calibrations=config.mfc_He_calibrations))
    mfcs.append(BronkhorstF201CV(serial_address=config.mfc_CO2_serial_address, serial_id=config.mfc_CO2_serial_id, calibrations=config.mfc_CO2_calibrations))
    mfcs.append(BronkhorstF201CV(serial_address=config.mfc_H2_serial_address, serial_id=config.mfc_H2_serial_id, calibrations=config.mfc_H2_calibrations))
    for mfc in mfcs:
        mfc.connect()
    return mfcs

def _check_flow_rates(mfcs:list[BronkhorstF201CV], flow_rates:list[float], plotter:DataCollectorPlotter|None=None):
    """
    Check if actual flow rates differ from expected less than 5%. Method expects equal number of items for mfcs and flow_rates objects with mutual correspondence of indicies. If plotter provided to method, it will stop plotting if actual flow rates differ from expected.

    parameters
    ----------
    mfcs:list[BronkhorstF201CV]
        list of mass flow controllers
    flow_rates:list[float]
        list of expected flow rates
    plotter:DataCollectorPlotter|None
        plotter which will be stopped if actual flow rates differ from expected ones

    raises
    ------
    Exception
        if actual flow rates differ from expected ones by more than 5%
    """
    for mfc, flow_rate in zip(mfcs, flow_rates):
        if flow_rate == 0:
            continue
        actual_flow_rate = mfc.get_flow_rate()
        if abs(actual_flow_rate - flow_rate) / flow_rate > 0.05:
            if plotter:
                plotter.stop()
            for mfc in mfcs:
                mfc.set_flow_rate(0)
            raise Exception("Actual gas flow rates differ from configuration!")

def _set_flow_rates(mfcs:list[BronkhorstF201CV], calibrations:list[int], flow_rates:list[float]):
    """
    Sets flow rates and calibrations of mass flow controllers

    parameters
    ----------
    mfcs:list[BronkhorstF201CV]
        list of mass flow controllers
    calibrations:list[int]
        list of calibration numbers with count started from 0
    flow_rates:list[float]
        list of flow rates
    """
    for mfc, calibration, flow_rate in zip(mfcs, calibrations, flow_rates):
        mfc.set_calibration(calibration_num=calibration)
        mfc.set_flow_rate(flow_rate)

def _heat_and_wait_until_temperature_reached(furnace:OwenTPM101, temperature:float):
    """
    Heat furnace to the required temperature and wait until this temperature is reached.

    parameters
    ----------
    furnace:OwenTPM101
        furnace controller object
    temperature:float
        required temperature
    """
    furnace.set_temperature_control(True)
    furnace.set_temperature(temperature)
    while True:
        current_temperature = furnace.get_temperature()
        if current_temperature >= temperature:
            break
        time.sleep(60)

def activate(args:argparse.Namespace):
    """
    Activate catalyst using parameters defined in configuration file, provided as argument. Configuration file is file with several variables created using python syntax. Use activation_config.py as an example. Method initializes furnace controller, mass flow controllers and connects to the devices. It sets mass flow controllers with proper calibrations and flow rates (corresponding valves must be opened prior this method is called). It waits 30 minutes for system to be purged with gases, heats furnace to activation temperature and holds it at that temperature for activation time. It then turns off heating, waits until furnace is cooled down and sets gas flow rates to the specified in configuration file values. NB: valves cannot be opened or closed automatically.
    """
    config_path = Path(args.config)
    process_config = _import_config(config_path)
    furnace_controller = _initialize_furnace_controller()
    mfcs = _initialize_mass_flow_controllers()
    _set_flow_rates(mfcs, process_config.calibrations, process_config.activation_flow_rates)
    plotter = DataCollectorPlotter(furnace_controller=furnace_controller, mass_flow_controllers=mfcs, gases=process_config.gases, chromatograph=None)
    plotter.start()
    # wait system to be purged with gases for 30 minutes
    time.sleep(30*60)
    _check_flow_rates(mfcs, process_config.activation_flow_rates, plotter=plotter)
    _heat_and_wait_until_temperature_reached(furnace_controller, process_config.activation_temperature)
    # dwell for activation duration time
    time.sleep(process_config.activation_duration*60)
    # turn off heating, wait until furnace is cooled down to post_temperature
    furnace_controller.set_temperature(0)
    furnace_controller.set_temperature_control(False)
    while True:
        current_temperature = furnace_controller.get_temperature()
        if current_temperature <= process_config.post_temperature:
            break
        time.sleep(60)
    # change gas flow rates to post activation values
    for mfc, flow_rate in zip(mfcs, process_config.post_flow_rates):
        mfc.set_flow_rate(flow_rate)
    input()
    plotter.stop()

def measure(args:argparse.Namespace):
    """
    Gather chromatograms at different measurement temperatures defined in a config file provided as an argument. Configuration file is a file with several variables defined using python syntax. Use measurement_config.py as an example of configuration. Method initializes devices and connects to them. It sets chromatograph method to 'purge', sets mass flow controller calibrations and flow rates. Heats furnace to the first measurement temperature and waits until target temperature is reached. Starts chromatograph purge, waits until purge is over and sets chromatograph method to the one specified in a config. Then, for each measurement temperature, method waits until chromatograph is ready for analysis, starts measurement, heats furnace to the next temperature. Finally, it turns off furnace and starts chromatograph cool down. During the process temperature, gas flow rates and chromatogram analysis start times are plotted. User has to hit enter after chromatograph cool down is started to stop plotter.
    """
    config_path = Path(args.config)
    process_config = _import_config(config_path)
    today = date.today()
    furnace = _initialize_furnace_controller()
    mfcs = _initialize_mass_flow_controllers()
    chromatograph = _initialize_chromatograph()
    plotter = DataCollectorPlotter(furnace_controller=furnace, mass_flow_controllers=mfcs, gases=process_config.gases, chromatograph=chromatograph)
    plotter.start()
    _set_flow_rates(mfcs, process_config.calibrations, process_config.flow_rates)
    chromatograph.set_method('purge')
    # wait 10 minutes to purge the system
    time.sleep(10*60)
    _check_flow_rates(mfcs, process_config.flow_rates)
    _heat_and_wait_until_temperature_reached(furnace, process_config.temperatures[0])
    # wait until chromatograph is ready for analysis, start chromatograph purge afterwards
    while True:
        chromatograph_is_ready = chromatograph.is_ready_for_analysis()
        if chromatograph_is_ready:
            break
        time.sleep(60)
    chromatograph.start_analysis()
    # wait until analysis is actually started
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is WorkingStatus.ANALYSIS:
            break
        time.sleep(60)
    # wait until chromatograph analysis is over, set passport values
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is not WorkingStatus.ANALYSIS:
            chromatograph.set_passport(name=f'{today.strftime("%Y%m%d")}_purge', volume=0.5, dilution=1, purpose=ChromatogramPurpose.ANALYSIS, operator=process_config.operator, column='HaesepN/NaX', lab_name='Inorganic Nanomaterials')
            break
        time.sleep(60)
    # wait until chromatograph starts to prepare itself, change instrumental method to corresponding instrumental method
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is WorkingStatus.PREPARATION or chromatograph_working_status is WorkingStatus.READY_FOR_ANALYSIS:
            chromatograph.set_method(process_config.chromatograph_method)
            break
        time.sleep(60)
    # for each temperature in measurement temperatures list:
        # wait until chromatograph is ready for analysis
        # read current furnace temperature
        # start chromatograph measurement
        # heat furnace to the next temperature
        # wait until temperature is reached
        # mark current time
        # wait until chromatograph analysis is over
        # set passport values
        # wait until isothermal dwell at measurement temperature is more than 30 minutes
    for temperature in process_config.temperatures[1:]:
        while True:
            if chromatograph.is_ready_for_analysis():
                break
            time.sleep(60)
        chromatogram_temperature = furnace.get_temperature()
        chromatograph.start_analysis()
        _heat_and_wait_until_temperature_reached(furnace, temperature)
        isothermal_start = time.time()
        while True:
            chromatograph_working_status = chromatograph.get_working_status()
            if chromatograph_working_status is not WorkingStatus.ANALYSIS:
                chromatograph.set_passport(name=f'{today.strftime("%Y%m%d")}_{process_config.sample_name}_{chromatogram_temperature:.1f}', volume=0.5, dilution=1, purpose=ChromatogramPurpose.ANALYSIS, operator=process_config.operator, column='HaesepN/NaX', lab_name='Inorganic Nanomaterials')
                break
            time.sleep(60)
        current_time = time.time()
        if current_time - isothermal_start < process_config.isothermal * 60:
            time.sleep(process_config.isothermal * 60 - (current_time - isothermal_start))
    # measure chromatogram at final temperature
    while True:
        if chromatograph.is_ready_for_analysis():
            break
        time.sleep(60)
    chromatogram_temperature = furnace.get_temperature()
    chromatograph.start_analysis()
    # wait until analysis is actually started
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is WorkingStatus.ANALYSIS:
            break
        time.sleep(60)
    # turn off heating
    furnace.set_temperature(0)
    furnace.set_temperature_control(False)
    # set final chromatogram passport values
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is not WorkingStatus.ANALYSIS:
            chromatograph.set_passport(name=f'{today.strftime("%Y%m%d")}_{process_config.sample_name}_{chromatogram_temperature:.1f}', volume=0.5, dilution=1, purpose=ChromatogramPurpose.ANALYSIS, operator=process_config.operator, column='HaesepN/NaX', lab_name='Inorganic Nanomaterials')
            break
        time.sleep(60)
    # start chromatograph cooldown
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is WorkingStatus.PREPARATION or chromatograph_working_status is WorkingStatus.READY_FOR_ANALYSIS:
            chromatograph.set_method('cooling')
            break
        time.sleep(60)
    input()
    plotter.stop()

def measure_init_conc(args:argparse.Namespace):
    """
    Method for measuring initial concentration of gas mixture needed for conversion calculation. Parameters of measurement must be in a configuration file provided as a parameter to this method. Method performs following actions:
        - sets gas flow rates to the values specified in config file
        - purges chromatograph prior to analysis
        - measures several chromatograms (number is defined in config file)
        - starts chromatograph cooldown
    Furnace temperature, gas flow rates and chromatograph analysis start times are plotted during the process. User has to hit enter after chromatograph started colling down to stop plotter.
    """
    config_path = Path(args.config)
    process_config = _import_config(config_path)
    today = date.today()
    furnace = _initialize_furnace_controller()
    mfcs = _initialize_mass_flow_controllers()
    chromatograph = _initialize_chromatograph()
    plotter = DataCollectorPlotter(furnace_controller=furnace, mass_flow_controllers=mfcs, gases=process_config.gases, chromatograph=chromatograph)
    plotter.start()
    chromatograph.set_method('purge')
    _set_flow_rates(mfcs, process_config.calibrations, process_config.flow_rates)
    # wait until chromatograph is ready to start analysis
    while True:
        chromatograph_is_ready = chromatograph.is_ready_for_analysis()
        if chromatograph_is_ready:
            break
        time.sleep(60)
    _check_flow_rates(mfcs, process_config.flow_rates)
    # purge chromatograph
    chromatograph.start_analysis()
    # wait until chromatograph analysis is actually started
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is WorkingStatus.ANALYSIS:
            break
        time.sleep(60)
    # wait until chromatograph analysis is over and write passport values
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is not WorkingStatus.ANALYSIS:
            chromatograph.set_passport(name=f'{today.strftime("%Y%m%d")}_purge', volume=0.5, dilution=1, purpose=ChromatogramPurpose.ANALYSIS, operator=process_config.operator, column='HaesepN/NaX', lab_name='Inorganic Nanomaterials')
            break
        time.sleep(60)
    # change chromatograph method to analysis method from config
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is WorkingStatus.PREPARATION or chromatograph_working_status is WorkingStatus.READY_FOR_ANALYSIS:
            chromatograph.set_method(process_config.chromatograph_method)
            break
        time.sleep(60)
    # gather series of chromatograms
    for i in range(process_config.measurements_number):
        while True:
            if chromatograph.is_ready_for_analysis():
                break
            time.sleep(60)
        chromatograph.start_analysis()
        while True:
            chromatograph_working_status = chromatograph.get_working_status()
            if chromatograph_working_status is WorkingStatus.ANALYSIS:
                break
            time.sleep(60)
        while True:
            chromatograph_working_status = chromatograph.get_working_status()
            if chromatograph_working_status is not WorkingStatus.ANALYSIS:
                chromatograph.set_passport(name=f'{today.strftime("%Y%m%d")}_{process_config.gases[0]}-{process_config.gases[1]}-{process_config.gases[2]}_{process_config.flow_rates[0]}-{process_config.flow_rates[1]}-{process_config.flow_rates[2]}_{i:02d}', volume=0.5, dilution=1, purpose=ChromatogramPurpose.ANALYSIS, operator=process_config.operator, column='HaesepN/NaX', lab_name='Inorganic Nanomaterials')
                break
            time.sleep(60)
    # wait until chromatograph method is finished and start cooldown
    while True:
        chromatograph_working_status = chromatograph.get_working_status()
        if chromatograph_working_status is WorkingStatus.PREPARATION or chromatograph_working_status is WorkingStatus.READY_FOR_ANALYSIS:
            chromatograph.set_method('cooling')
            break
        time.sleep(60)
    input()
    plotter.stop()

def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(required=True)

    calc_parser = subparsers.add_parser('calc', help='calculate conversion and selectivity vs. temperature')
    calc_parser.set_defaults(func=calculate)
    calc_parser.add_argument('input_data_path', metavar='input-data-path', help='path to directory with files from concentration measurement device')
    calc_parser.add_argument('initial_data_path', metavar='initial-data-path', help='path to file with data about initial composition of gas')
    calc_parser.add_argument('reaction', choices=['co-oxidation', 'co2-hydrogenation'], help='reaction for which to calculate data')
    calc_parser.add_argument('--conversion', action='store_true', help='calculate conversion for the specified reaction')
    calc_parser.add_argument('--selectivity', action='store_true', help='calculate selectivities for the specified reaction')
    calc_parser.add_argument('--output-data', default=None, help='path to directory to save calculated data')
    calc_parser.add_argument('--show-plot', action='store_true', help='whether to show data plot or not')
    calc_parser.add_argument('--output-plot', default=None, help='path to directory to save plot')
    calc_parser.add_argument('--products-basis', action='store_true', help='calculate conversion based on products concentration instead of reactants')
    calc_parser.add_argument('--sample-name', help='sample name will be added to results data files and as a title to the result plots')

    furnace_parser = subparsers.add_parser('furnace', help='control furnace')
    furnace_subparser = furnace_parser.add_subparsers(required=True)
    furnace_settemperature_parser = furnace_subparser.add_parser('set-temperature', help='set furnace temperature')
    furnace_settemperature_parser.set_defaults(func=furnace_set_temperature)
    furnace_settemperature_parser.add_argument('temperature', help='temperature in °C')
    furnace_printtemperature_parser = furnace_subparser.add_parser('print-temperature', help='print current temperature in °C')
    furnace_printtemperature_parser.set_defaults(func=furnace_print_temperature)

    chromatograph_parser = subparsers.add_parser('chromatograph', help='commands to control chromatograph')
    chromatograph_subparser = chromatograph_parser.add_subparsers(required=True)
    chromatograph_setmethod_parser = chromatograph_subparser.add_parser('set-method', help='set instrumental method of chromatograph')
    chromatograph_setmethod_parser.set_defaults(func=chromatograph_set_method)
    chromatograph_setmethod_parser.add_argument('method', help='method name')
    chromatograph_start_analysis_parser = chromatograph_subparser.add_parser('start-analysis', help='start analysis by chromatograph')
    chromatograph_start_analysis_parser.set_defaults(func=chromatograph_start_analysis)
    chromatograph_set_passport_parser = chromatograph_subparser.add_parser('set-passport', help='set chromatogram passport parameters (should be run after the analysis is over)')
    chromatograph_set_passport_parser.set_defaults(func=chromatograph_set_passport)
    chromatograph_set_passport_parser.add_argument('--name', required=True, help='name of chromatogram')
    chromatograph_set_passport_parser.add_argument('--volume', default=0.5, help='sample volume')
    chromatograph_set_passport_parser.add_argument('--dilution', default=1, help='sample dilution')
    chromatograph_set_passport_parser.add_argument('--purpose', default='analysis', choices=['analysis', 'graduation'], help='purpose of chromatogram')
    chromatograph_set_passport_parser.add_argument('--operator', required=True, help='operator\'s name')
    chromatograph_set_passport_parser.add_argument('--column', required=True, help='column\'s name')
    chromatograph_set_passport_parser.add_argument('--lab-name', default='Inorganic Nanomaterials', help='lab name')

    mfc_parser = subparsers.add_parser('mfc', help='commands to control mass flow controllers')
    mfc_subparser = mfc_parser.add_subparsers(required=True)
    mfc_set_flow_parser = mfc_subparser.add_parser('set-flow-rate', help='set gas flow rate')
    mfc_set_flow_parser.set_defaults(func=mfc_set_flow_rate)
    mfc_set_flow_parser.add_argument('--gas', required=True, choices=['He', 'CO2', 'O2', 'H2', 'CO', 'CH4'], help='which gas to set flow rate for')
    mfc_set_flow_parser.add_argument('--flow-rate', required=True, help='flow rate in nml/min')
    mfc_set_calibration_parser = mfc_subparser.add_parser('set-calibration', help='set calibration')
    mfc_set_calibration_parser.set_defaults(func=mfc_set_calibration)
    mfc_set_calibration_parser.add_argument('--gas', required=True, choices=['He', 'CO2', 'O2', 'H2', 'CO', 'CH4'], help='which gas to set calibration for')
    mfc_set_calibration_parser.add_argument('--calibration-number', required=True, help='number of calibration as written in calibraion document')
    mfc_print_flow_rate_parser = mfc_subparser.add_parser('print-flow-rate', help='print current flow rate')
    mfc_print_flow_rate_parser.set_defaults(func=mfc_print_flow_rate)
    mfc_print_flow_rate_parser.add_argument('--gas', required=True, choices=['He', 'CO2', 'O2', 'H2', 'CO', 'CH4'], help='which gas to print flow rate for')

    activation_parser = subparsers.add_parser('activate', help='activate catalyst using parameters provided in configuration file')
    activation_parser.set_defaults(func=activate)
    activation_parser.add_argument('--config', required=True, help='configuration file with activation parameters')

    measurement_parser = subparsers.add_parser('measure', help='gather chromatograms at different temperatures')
    measurement_parser.set_defaults(func=measure)
    measurement_parser.add_argument('--config', required=True, help='configuration file with measurement parameters')

    init_conc_parser = subparsers.add_parser('measure-init-concentration', help='gather several chromatograms consequently')
    init_conc_parser.set_defaults(func=measure_init_conc)
    init_conc_parser.add_argument('--config', required=True, help='path to configuration file with measurement parameters')

    args = parser.parse_args()
    args.func(args)



from pycatalicism.calc.co2hydrogenationcalculator import CO2HydrogenationCalculator
from pycatalicism.calc.rawdata import RawData
from pycatalicism.calc.conversion import Conversion
from pycatalicism.logging_decorator import Logging

class CO2HydrogenationProductsBasisCalculator(CO2HydrogenationCalculator):
    """
    Calculates CO2 conversion and CO, alkanes selectivity at different temperatures from parsed data for CO2 hydrogenation reaction based on products composition
    """

    @Logging
    def __init__(self):
        """
        Registers logger with the object which can be accessed via self.logger instance variable
        """
        super().__init__()

    def calculate_conversion(self, input_data:RawData) -> Conversion:
        """
        Calculates CO2 conversion at different temperatures for CO2 hydrogenation reaction based on products concentrations.

        CO2 conversion is calculated as:

        a = (SUM(nj * Cj) / C(CO2)i) * ((pf * ff * Ti) / (pi * fi * Tf))
        where
            nj - stoichiometric coefficient for jth product
                CO2 + H2 = CO + H2O, n = 1
                CO2 + 4H2 = CH4 + 2H2O, n = 1
                2CO2 + 7H2 = C2H6 + 4H2O, n = 2
                3CO2 + 10H2 = C3H8 + 6H2O, n = 3
                4CO2 + 13H2 = C4H10 + 8H2O, n = 4
                5CO2 + 16H2 = C5H12 + 10H2O, n = 5
            Cj - concentration of jth product in mol.%
            fi, ff - total gas flow rates before and after catalytic reactor, respectively, in m^3/s
            pi, pf - pressure of gas at point of total gas flow rate measurement before and after catalytic reactor, respectively, in Pa
            Ti, Tf - temperature of gas at point of total gas flow rate measurement before and after catalytic reactor, respectively, in K

        If flow rate measurement data is not provided, conversion is calculated based solely on concentrations and warning is logged to console in this case.

        parameters
        ----------
        input_data:RawData
            wrapper of parsed data containing product concentrations at different temperatures as well as initial CO2 concentration before reaction started

        returns
        -------
        conversion:Conversion
            wrapper with CO2 conversion at different temperatures data
        """
        self.logger.warning(f'Calculating conversion for CO2 hydrogenation reaction based on reaction products')
        temperatures = []
        alphas = []
        for temperature in input_data.get_temperatures():
            T_i = input_data.get_init_amb_temp()
            p_i = input_data.get_init_amb_pres()
            f_i = input_data.get_init_flow()
            T_f = input_data.get_fin_amb_temp(temperature)
            p_f = input_data.get_fin_amb_pres(temperature)
            f_f = input_data.get_fin_flow(temperature)
            C_CO2_i = input_data.get_init_conc('CO2')
            if T_i is None or p_i is None or f_i is None or T_f is None or p_f is None or f_f is None:
                self.logger.warning(f'No data about initial and final flow rate found. Calculating results based only on concentrations')
                T_i = 1
                p_i = 1
                f_i = 1
                T_f = 1
                p_f = 1
                f_f = 1
            product_sum = 0
            for compound in ['CO', 'CH4', 'C2H6', 'C3H8', 'i-C4H10', 'n-C4H10', 'i-C5H12', 'n-C5H12']:
                n = self._get_n(compound)
                product_sum = product_sum + n * input_data.get_conc(compound, temperature)
            alpha = (product_sum / C_CO2_i) * ((p_f * f_f * T_i) / (p_i * f_i * T_f))
            temperatures.append(temperature)
            alphas.append(alpha)
        conversion = Conversion(temperatures, alphas, input_data.get_sample_name())
        return conversion

    def _get_n(self, compound:str) -> int:
        """
        Get stoichiometry coefficient in CO2 hydrogenation reaction for specified compound.

        parameters
        ----------
        compound:str
            compound to calculate stoichiometry coefficient for

        returns
        -------
        n:int
            stoichiometry coefficient
        """
        n_str = compound[compound.find('C')+1]
        n = int(n_str) if n_str.isdecimal() else 1
        return n



import numpy as np

class Conversion():
    """
    Wrapper for conversion data storage. Conversion is stored as two parallel numpy.ndarrays of floats: temperature and conversion data.
    """

    def __init__(self, temperatures:list[float], alphas:list[float], sample_name:str|None):
        """
        Assign parameters to instance variables after conversion lists to numpy.ndarrays.

        parameters
        ----------
        temperatures:list[float]
            list of temperatures
        alphas:list[float]
            list of conversions
        sample_name:str|None
        """
        self.temperatures = np.array(temperatures)
        self.alphas = np.array(alphas)
        self.sample_name = sample_name

    def __str__(self) -> str:
        """
        Get string representation of covnersion vs. temperature data in a format:

        Sample<tab><sample-name><br>
        <br>
        Temperature<tab>Conversion<br>
        <temperature><tab><covnersion><br>
        ...

        returns
        -------
        string:str
            string representation of conversion vs. temperature data
        """
        string = f'Sample\t{self.sample_name}\n\nTemperature\tConversion\n'
        sorted_conversion = self.get_sorted()
        for temperature, alpha in zip(sorted_conversion.get_temperatures(), sorted_conversion.get_alphas()):
            string = string + f'{temperature}\t{alpha}\n'
        return string

    def get_sorted(self) -> 'Conversion':
        """
        Get conversion data sorted by temperature from lower to higher value

        returns
        -------
        conversion:Conversion
            wrapper of sorted conversion data
        """
        zipped_lists = zip(self.temperatures, self.alphas, strict=True)
        sorted_pairs = sorted(zipped_lists)
        tuples = zip(*sorted_pairs)
        sorted_temperatures, sorted_alphas = [list(tuple) for tuple in tuples]
        return Conversion(sorted_temperatures, sorted_alphas, self.sample_name)

    def get_temperatures(self) -> np.ndarray[float, np.dtype]:
        """
        Get temperatures as numpy.ndarray list

        returns
        -------
        temperatures:ndarray
            temperatures stored in this wrapper
        """
        return self.temperatures

    def get_alphas(self) -> np.ndarray[float, np.dtype]:
        """
        Get conversions as numpy.ndarray list

        returns
        -------
        conversions:ndarray
            conversions stored in this wrapper
        """
        return self.alphas



# parser type used to convert raw data from equipment for gas composition measurement to input data for calculator to calculate conversion/activity/selectivity vs. temperature
#
# chromatec-crystal-composition-copy-paste
#   raw data obtained via copy/paste of composition calculation results from chromatec analytics software
#   data must be in a format:
#   Температура<tab><temperature>
#   <br>
#   Название<tab>Время, мин<tab>Детектор<tab>Концентрация<tab>Ед, измерения<tab>Площадь<tab>Высота
#   <compound-name><tab><retention-time><tab><detector-name><tab><compound-concentration><tab><concentration-units><tab><peak-area><tab><peak-height>
#   [<br>
#   Темп. (газовые часы)<tab><flow-temperature>
#   Давление (газовые часы)<tab><flow-pressure>
#   Поток<tab><flow-rate>]
raw_data_parser_type = 'chromatec-crystal-composition-copy-paste'

# logging levels for different classes/modules
import logging
logging_levels = {
                    'ChromatecCrystalCompositionCopyPasteParser'    :   logging.INFO,
                    'CO2HydrogenationCalculator'                    :   logging.INFO,
                    'CO2HydrogenationProductsBasisCalculator'       :   logging.INFO,
                    'CO2HydrogenationExporter'                      :   logging.INFO,
                    'CO2HydrogenationPlotter'                       :   logging.INFO,
                    'COOxidationCalculator'                         :   logging.INFO,
                    'COOxidationExporter'                           :   logging.INFO,
                    'COOxidationPlotter'                            :   logging.INFO,
                    'RawData'                                       :   logging.INFO,
                    }

## chromatograph configuration ##
# slave id in control panel modbus configuration
control_panel_modbus_id = 1
# register with working status of chromatograph
working_status_input_address = 0
# register with step time in minutes
step_time_input_address = 18
# register with connection status
connection_status_input_address = 17
# register with current instrumental method
method_holding_address = 0
# register with chromatograph command
chromatograph_command_holding_address = 2
# register with control panel application command
application_command_holding_address = 3
# slave id in analytic modbus configuration
analytic_modbus_id = 2
# register with chromatogram nams
sample_name_holding_address = 0
# register with chromatogram purpose
chromatogram_purpose_holding_address = 15
# register with sample's volume
sample_volume_holding_address = 17
# register with sample's dilution
sample_dilution_holding_address = 21
# register with operator's name
operator_holding_address = 25
# register with column name
column_holding_address = 40
# register with lab name
lab_name_holding_address = 55
# list of methods with their order number at control panel software
methods = {
        '20220415_O2-N2-CO2-CO-C1,5HxAlkanes_2levels'	:		0,
        'co2-hydrogenation'								:		1,
        'cooling'										:		2,
        'co-oxidation'									:		3,
        'crm'											:		4,
        'Marusya method'								:		5,
        'NaX-conditioning'								:		6,
        'NaX-HaesepN-conditioning'						:		7,
        'purge'											:		8,
        'purge-overnight'								:		9,
        'zero'											:		10,
        'Водка-Маруся'									:		11,
        }

## mass flow controllers configuration ##
from pycatalicism.mass_flow_controller.bronkhorst_mfc_calibration import BronkhorstMFCCalibration
mfc_He_serial_address = 'COM3'
mfc_He_serial_id = 'M21212791C'
# NB: calibration ordering starts from 0 here, so, calibration for fluid1 has 0th number
mfc_He_calibrations = {
                    0	:	BronkhorstMFCCalibration(max_flow_rate=100, gas='He', p_in=3, p_out=1),
                    1	:	BronkhorstMFCCalibration(max_flow_rate=100, gas='He', p_in=32, p_out=30),
                    }
mfc_CO2_serial_address = 'COM4'
mfc_CO2_serial_id = 'M21212791E'
mfc_CO2_calibrations = {
                    0	:	BronkhorstMFCCalibration(max_flow_rate=100, gas='CO2', p_in=3, p_out=1),
                    1	:	BronkhorstMFCCalibration(max_flow_rate=100, gas='CO2', p_in=32, p_out=30),
                    2	:	BronkhorstMFCCalibration(max_flow_rate=100, gas='O2', p_in=3, p_out=1),
                    }
mfc_H2_serial_address = 'COM5'
mfc_H2_serial_id = 'M21212791D'
mfc_H2_calibrations = {
                    0	:	BronkhorstMFCCalibration(max_flow_rate=30, gas='CH4', p_in=3, p_out=1),
                    1	:	BronkhorstMFCCalibration(max_flow_rate=90, gas='CH4', p_in=32, p_out=30),
                    2	:	BronkhorstMFCCalibration(max_flow_rate=100, gas='H2', p_in=3, p_out=1),
                    3	:	BronkhorstMFCCalibration(max_flow_rate=100, gas='H2', p_in=32, p_out=30),
                    4	:	BronkhorstMFCCalibration(max_flow_rate=100, gas='CO', p_in=11, p_out=1),
                    }

## furnace configuration ##
import serial

# Furnace controller type
furnace_device_name = 'ÒÐÌ101' # <- this string is actually returned by the device \_O_/ although should be 'TPM101' according to owen protocol
# Furnace controller port name and corresponding port parameters (baudrate, bytesize, parity, stopbits) which must be the same as configured on controller device
furnace_port = 'COM6'
furnace_baudrate = 19200
furnace_bytesize = serial.EIGHTBITS
furnace_parity = serial.PARITY_NONE
furnace_stopbits = serial.STOPBITS_ONE
# Time in seconds to wait for the response from the device
furnace_timeout = 0.1
# Time in seconds to wait while message is sent to the device
furnace_write_timeout = None
# Enable/disable hardware flow control
furnace_rtscts = False
# Address of Owen TRM101 device
furnace_address = 0



import numpy as np

from pycatalicism.logging_decorator import Logging

class RawData():
    """
    Wrapper for imported data storage. Data are stored as numpy.ndarrays.
    """

    @Logging
    def __init__(self, temperatures:list[float]|np.ndarray[float,np.dtype], initial_concentrations:dict[str,float], concentrations:list[dict[str,float]]|np.ndarray[dict[str,float],np.dtype], initial_ambient_temperature:float|None=None, initial_ambient_pressure:float|None=None, initial_flow:float|None=None, final_ambient_temperatures:list[float]|np.ndarray[float,np.dtype]|None=None, final_ambient_pressures:list[float]|np.ndarray[float,np.dtype]|None=None, final_flows:list[float]|np.ndarray[float,np.dtype]|None=None, sample_name:str|None=None):
        """
        Registers logger with instance of this class which can be accessed via self.logger instance variable. Assigns parameters to instance variables converting lists to numpy.ndarray types.

        parameters
        ----------
        temperatures:list[float]|numpy.ndarray[float]
            list of temperatures at which gas composition measurements were done
        initial_concentrations:dict[str,float]
            dictionary of initial concentrations (i.e. before reaction started) of catalytic reaction compounds in a format {<compound>:<concentration>}
                compound:str
                    chemical formula of compound
                concentration:float
                    concentration in mol.%
        concentrations:list[dict[str,float]]|numpy.ndarray[dict[str,float]]
            list of compounds concentrations parrallel to temperatures list. Each list element is a dictionary in a format similar to initial_concentrations parameter
        initial_ambient_temperature:float|None (default:None)
            temperature of gas at the point of initial total flow rate measurement in °C
        initial_ambient_pressure:float|None (default:None)
            pressure of gas at the point of initial total flow rate measurement in Pa
        initial_flow:float|None (default:None)
            initial total flow rate in ml/min
        final_ambient_temperatures:list[float]|numpy.ndarray[float]|None (default:None)
            list of temperatures of gas at the point of total flow rate measurement in °C
        final_ambient_pressures:list[float]|numpy.ndarray[float]|None (default:None)
            list of pressures of gas at the point of total flow rate measurement in Pa
        final_flows:list[float]|numpy.ndarray[float]|None (default:None)
            list of total flow rates in ml/min
        sample_name:str|None (default:None)
            sample name which will be used as label for plotting
        """
        self.temperatures = np.array(temperatures)
        self.init_amb_temp = initial_ambient_temperature
        self.init_amb_pres = initial_ambient_pressure
        self.init_flow = initial_flow
        self.logger.debug(f'{final_ambient_temperatures = }')
        self.fin_amb_temps = None if final_ambient_temperatures is None else np.array(final_ambient_temperatures)
        self.fin_amb_pres = None if final_ambient_pressures is None else np.array(final_ambient_pressures)
        self.fin_flows = None if final_flows is None else np.array(final_flows)
        self.init_concs = initial_concentrations
        self.concs = np.array(concentrations)
        self.sample_name = sample_name

    def get_temperatures(self) -> np.ndarray[float, np.dtype]:
        """
        Get list of temperatures at which gas composition measurements were done

        returns
        -------
        temperatures:numpy.ndarray[float]
            list of temperatures
        """
        return self.temperatures

    def get_init_amb_temp(self) -> float|None:
        """
        Get temperature of gas at the point of initial total flow rate measurement in °C

        returns
        -------
        temperature:float|None
            temperature value in °C
        """
        return self.init_amb_temp

    def get_init_amb_pres(self) -> float|None:
        """
        Get pressure of gas at the point of initial total flow rate measurement in Pa

        returns
        -------
        pressure:float|None
            pressure in Pa
        """
        return self.init_amb_pres

    def get_init_flow(self) -> float|None:
        """
        Get initial total flow rate in ml/min

        returns
        -------
        flow_rate:float|None
            initial total flow rate in ml/min
        """
        return self.init_flow

    def get_fin_amb_temp(self, temperature:float) -> float|None:
        """
        Get temperature of gas at the point of total flow rate measurement during catalytic experiment at temperature provided as parameter to the method

        parameters
        ----------
        temperature:float
            temperature of catalytic experiment

        returns
        -------
        ambient_temperature:float|None
            temperature in °C
        """
        self.logger.debug(f'{self.fin_amb_temps = }')
        if self.fin_amb_temps is not None:
            return float(self.fin_amb_temps[self.temperatures == temperature])
        else:
            return None

    def get_fin_amb_pres(self, temperature:float) -> float|None:
        """
        Get pressure of gas at the point of total flow rate measurement during catalytic experiment at temperature provided as parameter to the method

        parameters
        ----------
        temperature:float
            temperature of catalytic reaction at the measurement moment

        returns
        -------
        ambient_pressure:float|None
            pressure in Pa
        """
        if self.fin_amb_pres is not None:
            return float(self.fin_amb_pres[self.temperatures == temperature])
        else:
            return None

    def get_fin_flow(self, temperature:float) -> float|None:
        """
        Get total gas flow rate during catalytic reaction at temperature provided as parameter to the method

        parameters
        ----------
        temperature:float
            temperature of catalytic reaction at the measurement moment

        returns
        -------
        flow_rate:float|None
            flow rate in ml/min
        """
        if self.fin_flows is not None:
            return float(self.fin_flows[self.temperatures == temperature])
        else:
            return None

    def get_init_conc(self, compound:str) -> float:
        """
        Get initial concentration of compound

        parameters
        ----------
        compound:str
            compound to get initial concentration for

        returns
        -------
        initial_concentration:float
            initial concentration in mol.%
        """
        return self.init_concs[compound]

    def get_conc(self, compound:str, temperature:float) -> float:
        """
        Get concentration of compound at temperature of catalytic reaction provided as parameter to the method

        parameters
        ----------
        compound:str
            compound to get concentration for
        temperature:float
            temperature of catalytic reaction at which concentration was measured

        returns
        -------
        concentration:float
            concentration in mol.% or 0 if at specified temperature concentration for the compound is not found
        """
        self.logger.debug(f'{self.concs = }')
        self.logger.debug(f'{self.temperatures = }')
        self.logger.debug(f'{compound = }')
        self.logger.debug(f'{temperature = }')
        try:
            conc = self.concs[self.temperatures == temperature][0][compound]
        except KeyError:
            self.logger.warning(f'Did not find concentration for "{compound}" at "{temperature}". Returning zero')
            conc = 0
        return conc

    def get_sample_name(self) -> str|None:
        """
        Get sample name

        returns
        -------
        sample_name:str|None
            name of sample
        """
        return self.sample_name



"""
Module for instantiating and configuring of logger
"""

import logging

from pycatalicism.furnace.logging_config import logging_levels

def get_logger(name:str) -> logging.Logger:
    """
    Get logger with corresponding name configured to log to stdout.

    parameters
    ----------
    name:str
        name of returned logger

    returns
    -------
    logger:logging.Logger
        configured logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging_levels[name])
    logger.propagate = False

    ch = logging.StreamHandler()
    ch.setLevel(logging_levels[name])

    formatter = logging.Formatter(fmt='[%(asctime)s] %(name)s.%(funcName)s: %(levelname)s: %(message)s', datefmt='%d.%m.%Y %H:%M:%S')

    ch.setFormatter(formatter)

    logger.addHandler(ch)

    return logger



import struct

import pycatalicism.chromatograph.chromatograph_logging as chromatograph_logging

_logger = chromatograph_logging.get_logger(__name__)

def bytes_to_float(response_bytes:list[int]) -> float:
    """
    Converts bytes received from chromatograph to float (8 byte double).

    parameters
    ----------
    response_bytes:list[int]
        bytes received from chromatograph

    returns
    -------
    float_num:float
        converted float value
    """
    _logger.debug(f'Converting bytes: {response_bytes} to float')
    float_bytes = b''
    for b in response_bytes:
        float_bytes += b.to_bytes(2, 'little')
    _logger.log(5, f'String bytes: {float_bytes = }')
    float_num, = struct.unpack('<d', float_bytes)
    _logger.log(5, f'{float_num = }')
    return float_num

def bytes_to_string(response_bytes:list[int]) -> str:
    """
    Converts bytes received from chromatograph to string.

    parameters
    ----------
    response_bytes:list[int]
        bytes received from chromatograph

    returns
    -------
    string:str
        decoded string
    """
    _logger.debug(f'Converting bytes: {response_bytes} to string')
    string = b''
    for b in response_bytes:
        string += b.to_bytes(2, 'big')
    _logger.log(5, f'String bytes: {string = }')
    string = string.decode().rstrip('\x00')
    _logger.log(5, f'{string = }')
    return string

def bytes_to_int(response_bytes:list[int]) -> int:
    """
    Converts bytes received from chromatograph to integer.

    parameters
    ----------
    response_bytes:list[int]
        bytes received from chromatograph

    returns
    -------
    integer:int
        decoded integer
    """
    _logger.debug(f'Converting bytes: {response_bytes} to int')
    integer = response_bytes[0]
    _logger.log(5, f'{integer = }')
    return integer

def string_to_bytes(string:str) -> tuple[int]:
    """
    Encodes string into bytes to be sent to chromatograph. NB: if string is more than 30 chars long, it will be concatenated to 30 chars due to the modbus limitations. A warning is logged in this case.

    parameters
    ----------
    string:str
        string to encode

    returns
    -------
    message:tuple[int]
        bytes to sent to chromatograph
    """
    _logger.debug(f'Converting string {string} to bytes')
    string = string.ljust(30)
    if len(string) > 30:
        _logger.warning(f'String cannot be > 30 chars long due to modbus limitation. Will be cut to 30 chars')
        string = string[0:30]
    string_bytes = bytes(string.encode())
    _logger.log(5, f'{string_bytes = }')
    message = struct.unpack('>'+'H'*15, string_bytes)
    _logger.log(5, f'{message = }')
    return message

def double_to_bytes(double:float) -> tuple[int]:
    """
    Encodes double value into bytes to be sent to chromatograph.

    parameters
    ----------
    double:float
        value to encode

    returns
    -------
    message:tuple[int]
        bytes to sent to chromatograph
    """
    _logger.debug(f'Converting double {double} to bytes')
    double_bytes = struct.pack('<d', double)
    _logger.log(5, f'{double_bytes = }')
    message = struct.unpack('<HHHH', double_bytes)
    _logger.log(5, f'{message = }')
    return message

def int_to_bytes(integer:int) -> list[int]:
    """
    Convert integer to bytes to be sent to chromatograph.

    parameters
    ----------
    integer:int
        value to encode

    returns
    -------
    message:list[int]
        bytes to sent to chromatograph
    """
    _logger.debug(f'Converting integer {integer} to bytes')
    integer_bytes = [integer]
    _logger.log(5, f'{integer_bytes = }')
    return integer_bytes



import multiprocessing
import multiprocessing.connection
import time
import threading

from pycatalicism.furnace.owen_tmp101 import OwenTPM101
from pycatalicism.mass_flow_controller.bronkhorst_f201cv import BronkhorstF201CV
from pycatalicism.chromatograph.chromatec_crystal_5000 import ChromatecCrystal5000
from pycatalicism.chromatograph.chromatograph_exceptions import ChromatographStateException
from pycatalicism.plotters.non_blocking_plotter import NonBlockingPlotter
from pycatalicism.plotters.point import Point

class DataCollectorPlotter(threading.Thread):
    """
    Class for plotting data from activation and activity measurement experiments.
    """

    def __init__(self, furnace_controller:OwenTPM101, mass_flow_controllers:list[BronkhorstF201CV], gases:list[str], chromatograph:ChromatecCrystal5000|None):
        """
        Initialize base class, instance variables and start non blocking plotter in a separate process.

        parameters
        ----------
        furnace_controller:OwenTPM101
            furnace controller to get information about current temperature
        mass_flow_controllers:list[BronkhorstF201CV]
            list of mass flow controllers to get information about current flow rates
        gases:list[str]
            list of gases used for process, which will be added as legend to plot
        chromatograph:ChromatecCrystal5000|None
            chromatograph to get information about analysis start time
        """
        super().__init__(daemon=False)
        self._furnace_controller = furnace_controller
        self._mfcs = mass_flow_controllers
        self._gases = gases
        self._chromatograph = chromatograph
        self._collector_pipe, self._plotter_pipe = multiprocessing.Pipe()
        self._plotter = NonBlockingPlotter()
        self._plotter_process = multiprocessing.Process(target=self._plotter, args=(self._plotter_pipe,), daemon=False)
        self._plotter_process.start()

    def run(self):
        """
        This method is invoked when thread is started. While thread is running get temperature, chromatograph and flow rate data and send these data through multiprocessing pipe to non blocking plotter. Send None to plotter, when thread is not running.
        """
        self._running = True
        self._start_time = time.time()
        while self._running:
            temperature_point, chromatograph_point, flow_rate_points = self._collect_data()
            self._collector_pipe.send((temperature_point, chromatograph_point, flow_rate_points))
            time.sleep(10)
        self._collector_pipe.send((None, None, None))

    def stop(self):
        """
        Stop this thread.
        """
        self._running = False

    def _collect_data(self) -> tuple[Point, Point|None, list[Point]]:
        """
        Get data about current temperature, time from the beginning of the chromatographic analysis and current gas flow rates.

        returns
        -------
        temperature_point:Point
            time and furnace temperature
        chromatograph_point:Point|None
            time from the beginning of the chromatographic analysis or None if chromatograph is not used or if current chromatograph working status is not analysis
        flow_rate_points:list[Point]
            list of times and flow rates for all mass flow controllers
        """
        t = (time.time() - self._start_time) / 60.0
        T = self._furnace_controller.get_temperature()
        temperature_point = Point(x=t, y=T, label='temperature')
        chromatograph_point = None
        if self._chromatograph is not None:
            try:
                t = round((time.time() - self._start_time) / 60.0, 2) - round(self._chromatograph.get_analysis_time(), 2)
                chromatograph_point = Point(x=t, y=None, label='chromatograms')
            except ChromatographStateException:
                pass
        flow_rate_points = []
        for mfc, gas in zip(self._mfcs, self._gases):
            t = (time.time() - self._start_time) / 60.0
            fr = mfc.get_flow_rate()
            flow_rate_point = Point(x=t, y=fr, label=gas)
            flow_rate_points.append(flow_rate_point)
        return (temperature_point, chromatograph_point, flow_rate_points)



from pathlib import Path

from pycatalicism.calc import calculator_factory
from pycatalicism.calc import parser_factory
from pycatalicism.calc import exporter_factory
from pycatalicism.calc import plotter_factory
from pycatalicism.calc.conversion import Conversion
from pycatalicism.calc.selectivity import Selectivity
from pycatalicism.calc.calculatorexception import CalculatorException

"""
Main interface for calculating conversion, activity and selectivity of catalysts. Users of this module should use calculate method to perform all calculation tasks.
"""


def _print_results(conversion:Conversion|None, selectivity:Selectivity|None):
    """
    Print results for conversion and/or selectivities vs. temperature to console.

    parameters
    ----------
    conversion:Conversion
        Reactant conversions at different temperatures data wrapper
    selectivity:Selectivity
        Catalyst selectivities for different product components at different temperatures data wrapper
    """
    if conversion:
        print(conversion)
    if selectivity:
        print(selectivity)

def calculate(input_data_path:str, initial_data_path:str, reaction:str, parser_type:str, calculate_conversion:bool, calculate_selectivity:bool, products_basis:bool=False, output_data_path:str|None=None, show_plot:bool=False, output_plot_path:str|None=None, sample_name:str|None=None):
    """
    Main interface to module. Parses input data from equipment capable of measuring composition and, ideally, initial and final gas total flow rate. Calculates conversion and/or selectivity data from input data. Prints results to console. If output_data_path was provided exports results. If show_plot is True, shows resulting plots. If output_plot_path was provided, exports corresponding plots.

    parameters
    ----------
    input_data_path:str
        Path to directory with input data files
    initial_data_path:str
        Path to file with gas composition data without catalyst (i.e. no reaction occured)
    reaction:str {co-oxidation|co2-hydrogenation}
        Chemical reaction to calculate data for
    parser_type:str {chromatec-crystal-composition-copy-paste}
        Parser type to use for parsing input data
    calculate_conversion:bool
        Whether to calculate conversion
    calculate_selectivity:bool
        Whether to calculate selectivity
    products_bases:bool (default:False)
        If True, calculate conversion based on products concentrations
    output_data_path:str|None {default:None}
        Path to directory to export results in text format
    show_plot:bool {default:False}
        Whether to show resulting plot
    output_plot_path:str|None {default:None}
        Path to directory to export resulting plot
    sample_name:str|None (default:None)
        Sample name which will be appended at the beginning of resulting data and as a plot title on resulting plot
    """
    if not (calculate_conversion or calculate_selectivity):
        raise CalculatorException('Nothing to calculate')
    calculator = calculator_factory.get_calculator(reaction, products_basis)
    parser = parser_factory.get_parser(parser_type)
    input_data = parser.parse_data(Path(input_data_path).resolve(), Path(initial_data_path).resolve(), sample_name)
    conversion = None
    selectivity = None
    if calculate_conversion:
        conversion = calculator.calculate_conversion(input_data)
    if calculate_selectivity:
        selectivity = calculator.calculate_selectivity(input_data)
    _print_results(conversion, selectivity)
    if output_data_path is not None:
        exporter = exporter_factory.get_exporter(reaction)
        exporter.export(Path(output_data_path).resolve(), conversion, selectivity)
    if show_plot or (output_plot_path is not None):
        plotter = plotter_factory.get_plotter(reaction)
        path = None if output_plot_path is None else Path(output_plot_path).resolve()
        plotter.plot(conversion, selectivity, show_plot, path, sample_name)



from pycatalicism.plotters.point import Point

class Data():
    """
    Wrapper class for storage of 2-D array of numbers and label. This class is used for non blocking plotting.
    """

    def __init__(self, label:str):
        """
        Instantiates instance variables.

        parameters
        ----------
        label:str
            label to associate with the data
        """
        self._label = label
        self._x = []
        self._y = []

    def add_point(self, point:Point|None):
        """
        Add one point to this data. If point is None, nothing is added. Also, if x value of point is already present in this data, nothing happens. Y value of the point can be None. In this case it is not appended to the data y array.

        parameters
        ----------
        point:Point|None
            point with x and y float values
        """
        if point is None:
            return
        x = point.get_x()
        y = point.get_y()
        if x not in self._x:
            self._x.append(x)
            if y is not None:
                self._y.append(y)

    def get_x(self) -> list[float]:
        """
        Get x values stored in this data object.

        returns
        -------
        x:list[float]
            list of x values
        """
        return self._x

    def get_y(self) -> list[float]:
        """
        Get y values stored in this data object.

        returns
        -------
        y:list[float]
            list of y values
        """
        return self._y

    def get_label(self) -> str:
        """
        Get label associated with this data object.

        returns
        -------
        label:str
            label associated with this data object
        """
        return self._label



import multiprocessing.connection

import matplotlib.pyplot as plt
import matplotlib
import matplotlib.axes

from pycatalicism.plotters.data import Data

class NonBlockingPlotter():
    """
    Non blocking plotter that should be run in a separate process due to the incompatibility of matplotlib with multithreading. It will draw data from instance variables every minute after the process was started. Data must be sent over multiprocessing pipe.
    """

    def __init__(self):
        """
        Initialize instance variables.
        """
        self._temperatures = Data(label='temperature')
        self._chromatograms = None
        self._flow_rates = None

    def __call__(self, pipe:multiprocessing.connection.Connection):
        """
        Create figure and axeses, setup them, start timer and show canvas.

        parameters
        ----------
        pipe:multiprocessing.connection.Connection
            pipe through which data are get from data collector
        """
        self._pipe = pipe
        self._fig = plt.figure()
        self._left_ax = self._fig.add_axes([0.1, 0.1, 0.6, 0.85]) # left, bottom, right, top
        self._setup_left_ax(self._left_ax)
        self._right_ax = self._left_ax.twinx()
        self._setup_right_ax(self._right_ax)
        timer = self._fig.canvas.new_timer(interval=60000)
        timer.add_callback(self._call_back)
        timer.start()
        plt.show()

    def _call_back(self) -> bool:
        """
        Collect data through multiprocessing pipe, append them to instance variables and add corresponding dots on plot canvas. This function is called by the matplotlib timer. Data must be sent in a 3-element tuple: (temperature_point, chromatogram_point, list-of-flow-rate-points). Each data point is wrapped as Point object.

        returns
        -------
        timer_is_running:bool
            True if timer is needed to be run, False if it must be stopped
        """
        while self._pipe.poll():
            data = self._pipe.recv()
            temperature_point = data[0]
            chromatogram_point = data[1]
            flow_rate_points = data[2]
            if temperature_point is None:
                return False
            else:
                self._temperatures.add_point(temperature_point)
                if chromatogram_point is not None:
                    if self._chromatograms is None:
                        self._chromatograms = Data(label='chromatograms')
                    self._chromatograms.add_point(chromatogram_point)
                if self._flow_rates is None:
                    self._flow_rates = []
                    for flow_rate_point in flow_rate_points:
                        self._flow_rates.append(Data(label=flow_rate_point.get_label()))
                for flow_rate_point, flow_rate in zip(flow_rate_points, self._flow_rates):
                    flow_rate.add_point(flow_rate_point)
                self._left_ax.clear()
                self._right_ax.clear()
                self._setup_left_ax(self._left_ax)
                self._setup_right_ax(self._right_ax)
                lines = []
                line, = self._left_ax.plot(self._temperatures.get_x(), self._temperatures.get_y(), color='red', linewidth=1, linestyle=':', label=self._temperatures.get_label())
                lines.append(line)
                if self._chromatograms is not None:
                    line = self._left_ax.vlines(self._chromatograms.get_x(), self._left_ax.get_ylim()[0], self._left_ax.get_ylim()[1], colors=['#000000'], linewidth=1, label=self._chromatograms.get_label())
                    lines.append(line)
                for flow_rate in self._flow_rates:
                    line, = self._right_ax.plot(flow_rate.get_x(), flow_rate.get_y(), linewidth=1, linestyle=':', label=flow_rate.get_label())
                    lines.append(line)
                self._left_ax.legend(handles=lines, bbox_to_anchor=(1.1, 1), loc='upper left', borderaxespad=0)
        self._fig.canvas.draw()
        return True

    def _setup_left_ax(self, left_ax:matplotlib.axes.Axes):
        """
        Setup left, temperature, axes. Set x and y labels

        parameters
        ----------
        left_ax:matplotlib.axes.Axes
            axes to setup
        """
        left_ax.set_xlabel('Time, min')
        left_ax.set_ylabel('Temperature, °C')

    def _setup_right_ax(self, right_ax:matplotlib.axes.Axes):
        """
        Setup right, flow rates, axes. Sets y label.

        parameters
        ----------
        right_ax:matplotlib.axes.Axes
            axes to setup
        """
        right_ax.set_ylabel('Flow rate, n.ml/min')



from enum import Enum
import time

from pymodbus.client.sync import ModbusTcpClient

import pycatalicism.chromatograph.chromatograph_logging as chromatograph_logging
import pycatalicism.chromatograph.modbus_converter as convert
from pycatalicism.chromatograph.chromatograph_exceptions import ChromatographException

class WorkingStatus(Enum):
    """
    Working status of chromatograph and corresponding values in modbus protocol
    """
    NULL = 0
    PREPARATION = 1
    READY_FOR_ANALYSIS = 4
    PURGING = 6
    COOLING = 8
    ANALYSIS = 9

class ConnectionStatus(Enum):
    """
    Connection status of chromatograph and corresponding values in modbus protocol
    """
    CP_ON_CONNECTED = 7
    CP_ON_CONNECTING = 3 # not sure about exact meaning, however it is sometimes returned right after control panel is up, but chromatograph is not connected
    CP_ON_NOT_CONNECTED = 1
    CP_OFF_NOT_CONNECTED = 0

class ChromatographCommand(Enum):
    """
    Commands for chromatograph control and corresponding values in modbus protocol
    """
    CONNECT_CHROMATOGRAPH = 1
    START_ANALYSIS = 6

class ApplicationCommand(Enum):
    """
    Commands for application control and corresponding values in modbus protocol
    """
    START_CONTROL_PANEL = 1

class ChromatecControlPanelModbus():
    """
    Class represents simplified version of modbus protocol for connection with chromatec control panel modbus slave.
    """

    def __init__(self, modbus_id:int, working_status_input_address:int, connection_status_input_address:int, step_time_input_address:int, method_holding_address:int, chromatograph_command_holding_address:int, application_command_holding_address:int, request_trials:int=3):
        """
        Initializes instance private variables, creates modbus client and registers logger.

        parameters
        ----------
        modbus_id:int
            modbus slave id of control panel software
        working_status_input_address:int
            modbus address of current chromatograph status (analysis, purging, preparing etc.)
        connection_status_input_address:int
            modbus address for chromatograph and its applications connection status
        step_time_input_address:int
            modbus address for time from the start of analysis
        method_holding_address:int
            modbus address for setting instrumental methods
        chromatograph_command_holding_address:int
            modbus address for chromatograph commands (see chromatec modbus manual for details)
        application_command_holding_address:int
            modbus address for application commands (see chromatec modbus manual for details)
        """
        self._modbus_id = modbus_id
        self._request_trials = request_trials
        self._working_status_input_address = working_status_input_address
        self._connection_status_input_address = connection_status_input_address
        self._step_time_input_address = step_time_input_address
        self._method_holding_address = method_holding_address
        self._chromatograph_command_holding_address = chromatograph_command_holding_address
        self._application_command_holding_address = application_command_holding_address
        self._modbus_client = ModbusTcpClient()
        self._logger = chromatograph_logging.get_logger(self.__class__.__name__)

    def get_current_working_status(self) -> WorkingStatus:
        """
        Get working status of chromatograph, i.e. analysis, purging, preparing etc.

        returns
        -------
        current_status:WorkingStatus
            one of the constants defined in WorkingStatus enum
        """
        count = 0
        while True:
            self._logger.debug(f'Getting current working status of chromatograph. Trial #{count}')
            try:
                response = self._modbus_client.read_input_registers(address=self._working_status_input_address, count=2, unit=self._modbus_id)
                response_registers = response.registers
                break
            except AttributeError:
                if count == self._request_trials-1:
                    raise ChromatographException('Cannot get working status!')
                else:
                    count += 1
                    self._logger.warning(f'Error during communication with control panel. Trying to reconnect. Trial #{count}')
                    time.sleep(1)
        current_status_id = convert.bytes_to_int(response_registers)
        current_status = WorkingStatus(current_status_id)
        self._logger.log(5, f'{current_status = }')
        return current_status

    def get_step_time(self) -> float:
        """
        Get time passed from the beginning of the current step in minutes.

        returns
        -------
        step_time:float
            time from the beginning of the current step in minutes

        raises
        ------
        ChromatographException
            if error occured during communication with control panel
        """
        count = 0
        while True:
            self._logger.debug(f'Getting analysis time. Trial #{count}')
            try:
                response = self._modbus_client.read_input_registers(address=self._step_time_input_address, count=4, unit=self._modbus_id)
                response_registers = response.registers
                self._logger.log(5, f'{response_registers = }')
                break
            except AttributeError:
                if count == self._request_trials-1:
                    raise ChromatographException('Cannot get analysis time!')
                else:
                    count += 1
                    self._logger.warning(f'Error during communication with control panel. Trying to reconnect. Trial #{count}')
                    time.sleep(1)
        step_time = convert.bytes_to_float(response_registers)
        self._logger.debug(f'{step_time = }')
        return step_time

    def get_connection_status(self) -> ConnectionStatus:
        """
        Get status of connection with chromatograph and control panel modbus slave.

        returns
        -------
        connection_status:ConnectionStatus
            one of the constants defined in ConnectionStatus enum
        """
        count = 0
        while True:
            self._logger.debug(f'Getting chromatograph and control panel connection status. Trial #{count}')
            try:
                response = self._modbus_client.read_input_registers(address=self._connection_status_input_address, count=1, unit=self._modbus_id)
                response_registers = response.registers
                break
            except AttributeError:
                if count == self._request_trials-1:
                    raise ChromatographException('Cannot get connection status!')
                else:
                    count += 1
                    self._logger.warning(f'Error during communication with control panel. Trying to reconnect. Trial #{count}')
                    time.sleep(1)
        connection_status = ConnectionStatus(response_registers[0])
        self._logger.log(5, f'{connection_status = }')
        return connection_status

    def set_instrument_method(self, method_id:int):
        """
        Sets instrumental method to specified method id.

        parameters
        ----------
        method_id:int
            sequential number of instrumental method
        """
        self._logger.debug(f'Setting instrumental method to {method_id}')
        self._modbus_client.write_registers(address=self._method_holding_address, values=[method_id], unit=self._modbus_id)

    def send_chromatograph_command(self, command:ChromatographCommand):
        """
        Send command to chromatograph.

        parameters
        ----------
        command:ChromatographCommand
            one of the constants defined in ChromatographCommand enum
        """
        self._logger.debug(f'Sending command to chromatograph: {command}')
        self._modbus_client.write_registers(address=self._chromatograph_command_holding_address, values=[command.value], unit=self._modbus_id)

    def send_application_command(self, command:ApplicationCommand):
        """
        Send command to control panel application.

        parameters
        ----------
        command:ApplicationCommand
            one of the constants defined in ApplicationCommand enum
        """
        self._logger.debug(f'Sending command to control panel: {command}')
        self._modbus_client.write_registers(address=self._application_command_holding_address, values=[command.value], unit=self._modbus_id)



import logging

# Dictionary of logging levels of corresponding classes
logging_levels = {
                'OwenProtocol'	:	logging.INFO,
                'OwenTPM101'	:	logging.INFO,
                }
